------------------------------------------------------------------------------
--                                                                          --
--                          GNATCHECK COMPONENTS                            --
--                                                                          --
--                   G N A T C H E C K . P R O J E C T S                    --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                       Copyright (C) 2013, AdaCore                        --
--                                                                          --
-- GNATCHECK  is  free  software;  you can redistribute it and/or modify it --
-- under terms of the  GNU  General Public License as published by the Free --
-- Software Foundation;  either version 2, or ( at your option)  any  later --
-- version.  GNATCHECK  is  distributed in the hope that it will be useful, --
-- but  WITHOUT  ANY  WARRANTY;   without  even  the  implied  warranty  of --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details.  You should have received a copy of the --
-- GNU  General Public License distributed with GNAT; see file  COPYING. If --
-- not,  write to the  Free Software Foundation,  51 Franklin Street, Fifth --
-- Floor, Boston, MA 02110-1301, USA.                                       --
--                                                                          --
-- GNATCHECK is maintained by AdaCore (http://www.adacore.com).             --
--                                                                          --
------------------------------------------------------------------------------

with Ada.Strings.Fixed;

with GNAT.Command_Line;          use GNAT.Command_Line;
with GNAT.OS_Lib;                use GNAT.OS_Lib;

with GNATCOLL.Projects;          use GNATCOLL.Projects;

with ASIS_UL.Common;             use ASIS_UL.Common;
with ASIS_UL.Debug;              use ASIS_UL.Debug;
with ASIS_UL.Options;            use ASIS_UL.Options;
with ASIS_UL.Output;             use ASIS_UL.Output;

with Gnatcheck.Diagnoses_Old;
with Gnatcheck.Options;          use Gnatcheck.Options;
with Gnatcheck.Output;           use Gnatcheck.Output;
with Gnatcheck.Rules.Rule_Table; use Gnatcheck.Rules.Rule_Table;

package body Gnatcheck.Projects is

   ------------------------
   -- Extract_Tool_Xvars --
   ------------------------

   procedure Extract_Tool_Xvars (My_Project : in out Gnatcheck_Project_Type) is
      Success : Boolean;
      Proj    : constant Project_Type := Root_Project (My_Project);

      Attr : constant Attribute_Pkg_List :=
        Build (Tool_Package_Name (My_Project), "Default_Switches");

      Tool_Switches : String_List_Access;

      Proj_Args_Parser : Opt_Parser;
   begin
      if Has_Attribute (Proj, Attr) then
         Tool_Switches := Attribute_Value (Proj, Attr, "ada");

         Initialize_Option_Scan
           (Parser                   => Proj_Args_Parser,
            Command_Line             => Tool_Switches,
            Switch_Char              => '-',
            Stop_At_First_Non_Switch => False);

         loop
            case GNAT.Command_Line.Getopt
              ("* from= rules "                   & --  rule options
               "v q t s? l m? a "                 &
               "X! "                        &
               "j! "                              &
               "d? "                              &

               "log "                             &
               "-include-file= "                  &

               "-old-report-format "              &

               "-no-column "                      &
               "-exemption "                      &
               "-show-rule "                      &
               "-test "                           &
               "o= "                              &
               "-write-rules= "                   &
      --       "-dump-code-standard=  "           &
               "gnat05",
               Parser => Proj_Args_Parser)
            is

               when ASCII.NUL =>
                  exit;
               when 'a' =>

                  if Full_Switch (Parser => Proj_Args_Parser) = "a" then
                     Process_RTL_Units := True;
                  end if;

               when 'd' =>

                  if Full_Switch (Parser => Proj_Args_Parser) = "d" then
                     Set_Debug_Options
                       (Parameter (Parser => Proj_Args_Parser));
                  end if;

               when 'f' =>

                  if Full_Switch (Parser => Proj_Args_Parser) = "from" then
                     Process_Rule_File
                       (Parameter (Parser => Proj_Args_Parser));

                     if not More_Then_One_Rule_File_Set then
                        Rule_File_Name :=
                          new String'(Parameter (Parser => Proj_Args_Parser));
                        More_Then_One_Rule_File_Set := True;
                     else
                        Free (Rule_File_Name);
                     end if;
                  end if;

               when 'g' =>

                  if Full_Switch (Parser => Proj_Args_Parser) = "gnat05" then
                     ASIS_UL.Options.ASIS_2005_Mode := True;
                  end if;

               when 'j' =>

                  begin
                     ASIS_UL.Options.Process_Num :=
                       Positive'Value (Parameter (Parser => Proj_Args_Parser));
                  exception
                     when Constraint_Error =>
                     Error ("Wrong parameter of '-j' option: " &
                            Parameter (Parser => Proj_Args_Parser));
                     raise Parameter_Error;
                  end;

               when 'l' =>

                  if Full_Switch (Parser => Proj_Args_Parser)  = "l" then
                     Gnatcheck.Options.Full_Source_Locations := True;
                  elsif Full_Switch (Parser => Proj_Args_Parser)  = "log" then
                     Log_Mode := True;
                     Set_Log_File;
                  end if;

               when 'm' =>

                  begin
                     Gnatcheck.Options.Max_Diagnoses :=
                       Natural'Value (Parameter (Parser => Proj_Args_Parser));

                     if Gnatcheck.Options.Max_Diagnoses > 1000 then
                        Error ("parameter of '-m' option too big, " &
                               "max allowed is 1000");
                        raise Parameter_Error;
                     end if;

                  exception
                     when Constraint_Error =>
                     Error ("Wrong parameter of '-m' option: " &
                            Parameter (Parser => Proj_Args_Parser));
                     raise Parameter_Error;
                  end;

               when 'o' =>

                  if Full_Switch (Parser => Proj_Args_Parser) = "o" then
                     Set_Report_File_Name
                       (Parameter (Parser => Proj_Args_Parser));
                  end if;

               when 'q' =>
                  Quiet_Mode := True;

               when 'r' =>
                  if Full_Switch (Parser => Proj_Args_Parser) = "rules" then
                     null;  --  there is no 'rules' section in a project file.
                  else
                     Error ("cannot process " &
                            Full_Switch (Parser => Proj_Args_Parser) &
                            " from a project file");
                     raise Parameter_Error;
                  end if;
               when 's' =>
                  Gnatcheck.Diagnoses_Old.Process_Report_File_Format_Parameter
                    (Parameter (Parser => Proj_Args_Parser), Success);

                  if not Success then
                     raise Parameter_Error;
                  end if;

               when 't' =>
                  Compute_Timing := True;

               when 'v' =>
                  Verbose_Mode := True;

                  Print_Version_Info (2004);

               when 'X' =>
                  ASIS_UL.Projects.Store_External_Variable
                    (Var => Parameter (Parser => Proj_Args_Parser),
                     Is_From_Command_Line => False);
               when '-' =>

                  if Full_Switch (Parser => Proj_Args_Parser) =
                     "-no-column"
                  then
                     Gnatcheck.Options.No_Column_Num_In_Diagnoses := True;
                  elsif Full_Switch (Parser => Proj_Args_Parser) =
                        "-exemption"
                  then
                     Gnatcheck.Options.Print_Exemption_Section := True;
                  elsif Full_Switch (Parser => Proj_Args_Parser) =
                        "-include-file"
                  then
                     Gnatcheck.Diagnoses_Old.Process_User_Filename
                       (Parameter (Parser => Proj_Args_Parser));
                  elsif Full_Switch (Parser => Proj_Args_Parser) =
                        "-show-rule"
                  then
                     Gnatcheck.Options.Mapping_Mode := True;
                  elsif Full_Switch (Parser => Proj_Args_Parser) =
                        "-test"
                  then
                     ASIS_UL.Options.Test_Mode := True;
                     ASIS_UL.Options.Warning_Mode := Quiet;
                  elsif Full_Switch (Parser => Proj_Args_Parser) =
                        "-old-report-format"
                  then
                     Gnatcheck.Options.Qualification_Report := False;
                  elsif Full_Switch (Parser => Proj_Args_Parser) =
                        "-write-rules"
                  then
                     Generate_Coding_Standard := True;
                     Set_Coding_Standard_File_Name
                       (Parameter (Parser => Proj_Args_Parser));
                  end if;

               when others =>
                  Process_Rule_Option
                    (Ada.Strings.Fixed.Trim (Full_Switch, Ada.Strings.Both));
                  --  We use the call to Trim here because there can be a rule
                  --  option in quotation marks
                  Individual_Rules_Set := True;
            end case;
         end loop;

      end if;
   end Extract_Tool_Xvars;

   -----------------------
   -- Tool_Package_Name --
   -----------------------

   function Tool_Package_Name
     (My_Project : Gnatcheck_Project_Type)
      return       String
   is
      pragma Unreferenced (My_Project);
   begin
      return "check";

   end Tool_Package_Name;

end Gnatcheck.Projects;
