------------------------------------------------------------------------------
--                                                                          --
--                            GNAT2XML COMPONENTS                           --
--                                                                          --
--                      G N A T 2 X M L . B U F F E R S                     --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                     Copyright (C) 2013, AdaCore, Inc.                    --
--                                                                          --
-- Gnat2xml is free software; you can redistribute it and/or modify it      --
-- under terms of the  GNU General Public License  as published by the Free --
-- Software Foundation;  either version 2,  or  (at your option)  any later --
-- version. Gnat2xml is distributed  in the hope  that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of MER-      --
-- CHANTABILITY or  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General  --
-- Public License for more details. You should have received a copy of the  --
-- GNU General Public License distributed with GNAT; see file COPYING. If   --
-- not, write to the Free Software Foundation, 59 Temple Place Suite 330,   --
-- Boston, MA 02111-1307, USA.                                              --
-- The gnat2xml tool was derived from the Avatox sources.                   --
------------------------------------------------------------------------------

pragma Ada_2012;

with Ada.Characters.Latin_1;
use Ada.Characters;

package body Ada_Trees.Buffers is

   procedure Append_New_Marker (Buf : in out Buffer; Name : Character);
   --  Append a new Marker onto To_Markers, making it point to 'point'

   procedure Maybe_Move_Marker (Buf : in out Buffer);
   --  If the first mark in From_Markers is at 'point', move it to the
   --  To_Markers array, adjusting its Position and Flag, to preserve the
   --  invariant.

   function Valid_From (Buf : Buffer) return String;
   --  Return the valid portion of Buf.From

   function Mx (Buf : Buffer; Mark : Marker) return Marker_Rec_Index;
   --  Subtract off the Unique_Id, for indexing into Markers

   procedure Initialize_Buffer (Buf : in out Buffer);

   procedure Append_New_Marker (Buf : in out Buffer; Name : Character) is
   begin
      Append
        (Buf.Markers,
         Marker_Rec'
           (Position => Last_Index (Buf.To) + 1,
           Flag      => Buf.To_Flag,
           Name      => Name));
      Append
        (Buf.To_Markers,
         Marker (Last_Index (Buf.Markers)) + Buf.Unique_Id);
   end Append_New_Marker;

   function At_End (Buf : Buffer) return Boolean is
   begin
      return Buf.From_First > Last_Index (Buf.From);
   end At_End;

   function At_Point (Buf : Buffer; Mark : Marker) return Boolean is
      Rec : constant Marker_Rec := Buf.Markers (Mx (Buf, Mark));

   begin
      return
        Rec.Flag = Buf.To_Flag
        and then Rec.Position = Last_Index (Buf.To) + 1;
   end At_Point;

   function Char_At (Buf : Buffer; Mark : Marker) return Character is
      Rec : constant Marker_Rec := Buf.Markers (Mx (Buf, Mark));
      Result : Character;

   begin
      if Rec.Flag = Buf.To_Flag then
         if Rec.Position = Last_Index (Buf.To) + 1 then
            Result := Buf.From (Buf.From_First);

         else
            Result := Buf.To (Rec.Position);
         end if;

      else
         Result := Buf.From (Rec.Position);
      end if;
      pragma Assert (Result = Char_At (Buf, Position (Buf, Mark)));
      return Result;
   end Char_At;

   function Char_At (Buf : Buffer; Position : Positive) return Character is
   begin
      if Position <= Last_Index (Buf.To) then
         return Buf.To (Position);

      else
         return Buf.From (Position - Last_Index (Buf.To) + Buf.From_First - 1);
      end if;
   end Char_At;

   procedure Clear (Buf : in out Buffer) is
   begin
      pragma Debug (Validate (Buf, "Clear"));

      Clear (Buf.From);
      Clear (Buf.To);
      Clear (Buf.Markers);
      Clear (Buf.From_Markers);
      Clear (Buf.To_Markers);
      Initialize_Buffer (Buf);
   end Clear;

   function Cur (Buf : Buffer) return Character is
      Result : Character;

   begin
      if At_End (Buf) then
         Result := Latin_1.NUL;

      else
         Result := Buf.From (Buf.From_First);
      end if;
      return Result;
   end Cur;

   function Cur_Column (Buf : Buffer) return Positive is
   begin
      return Buf.Cur_Column;
   end Cur_Column;

   procedure Dump_Buffer (Buf : Buffer) is
      use Dbg_Out;

   begin
      if False then
         Put ("Buf = <<\1>>end Buf\n", To_String (Buf));
         Put ("Dbg = <<\1>>end Dbg\n", To_Debug_String (Buf));
      end if;
      Put
        ("To[\1..\2] = <<\3>>end To[\1..\2]\n",
         "1",
         Image (Last_Index (Buf.To)),
         To_Array (Buf.To));
      Put
        ("From[\1..\2] = <<\3>>end From[\1..\2]\n",
         Image (Buf.From_First),
         Image (Last_Index (Buf.From)),
         Valid_From (Buf));

      if False then
         Put ("To_Markers:\n");
         Indent;

         for X in 1 .. Last_Index (Buf.To_Markers) loop
            Dump_Marker (Buf, Buf.To_Markers (X));
         end loop;
         Outdent;
         Put ("end To_Markers\n");

         Put
           ("From_Markers[\1..\2]:\n",
            Image (Integer (Buf.From_Markers_First)),
            Image (Integer (Last_Index (Buf.From_Markers))));
         Indent;

         for X in Buf.From_Markers_First .. Last_Index (Buf.From_Markers) loop
            Dump_Marker (Buf, Buf.From_Markers (X));
         end loop;
         Outdent;
         Put ("end From_Markers\n");
      end if;
   end Dump_Buffer;

   procedure Dump_Marker (Buf : Buffer; Mark : Marker) is
      use Dbg_Out;
      Rec : Marker_Rec renames Buf.Markers (Mx (Buf, Mark));
      Pos : constant Positive := Position (Buf, Mark);

   begin
      if Rec.Flag = Buf.To_Flag then
         Put ("to   ");

      else
         Put ("from ");
      end if;
      Put
        ("\1 --> \2 '\3'  \4\n",
         Image (Rec.Position),
         Image (Pos),
         (1 => Rec.Name),
         Image (Modular (Mark)));

      Put ("<<");

      for P in Pos - 30 .. Pos + 30 loop
         if P = Pos then
            Put_Char ('[');
         end if;
         if P in 1 .. Last_Position (Buf) then
            if Char_At (Buf, P) = NL then
               Put_Char ('$');

            else
               Put_Char (Char_At (Buf, P));
            end if;
         end if;
         if P = Pos then
            Put_Char (']');
         end if;
      end loop;
      Put (">>\n");
   end Dump_Marker;

   function Mark_LT (Buf : Buffer; M, N : Marker) return Boolean is
   begin
      return Position (Buf, M) < Position (Buf, N);
   end Mark_LT;

   function Mark_LE (Buf : Buffer; M, N : Marker) return Boolean is
   begin
      return Position (Buf, M) = Position (Buf, N);
   end Mark_LE;

   procedure Maybe_Move_Marker (Buf : in out Buffer) is
   begin
      --  We can't have two different Markers pointing to the same
      --  character (see Mark), so we don't need a while loop here.

      if Buf.From_Markers_First <= Last_Index (Buf.From_Markers) then
         declare
            Mark : constant Marker :=
              Buf.From_Markers (Buf.From_Markers_First);
            Rec : Marker_Rec renames Buf.Markers (Mx (Buf, Mark));

         begin
            pragma Assert (Rec.Position >= Buf.From_First);
            if Rec.Position = Buf.From_First then
               Append (Buf.To_Markers, Mark);
               Rec :=
                 (Flag    => not Rec.Flag,
                 Position => Last_Index (Buf.To) + 1,
                 Name     => Rec.Name);
               pragma Assert (Rec.Flag = Buf.To_Flag);
               Buf.From_Markers_First := Buf.From_Markers_First + 1;
               pragma Assert (At_Point (Buf, Mark));
            end if;
         end;
      end if;

      if Buf.From_Markers_First <= Last_Index (Buf.From_Markers) then
         declare
            Mark : constant Marker :=
              Buf.From_Markers (Buf.From_Markers_First);
            Rec : Marker_Rec renames Buf.Markers (Mx (Buf, Mark));

         begin
            pragma Assert
              (Rec.Position in
                 Buf.From_First + 1 .. Last_Index (Buf.From) + 1);
         end;

      else
         pragma Assert
           (Buf.From_Markers_First = Last_Index (Buf.From_Markers) + 1);
      end if;
   end Maybe_Move_Marker;

   procedure Move_Forward (Buf : in out Buffer) is
      C : Character;

   begin
      if not At_End (Buf) then
         C := Buf.From (Buf.From_First);
         Append (Buf.To, C);
         if C = NL then
            Buf.Cur_Column := 1;

         else
            Buf.Cur_Column := Buf.Cur_Column + 1;
         end if;
      end if;
      Buf.From_First := Buf.From_First + 1;

      Maybe_Move_Marker (Buf);
   end Move_Forward;

   procedure Initialize_Buffer (Buf : in out Buffer) is
   begin
      Buf.From_First         := 1;
      Buf.From_Markers_First := 1;
      Buf.Cur_Column         := 1;
      Buf.Unique_Id          := Cur_Unique_Id;
      Cur_Unique_Id          := Cur_Unique_Id + 1_000_000; -- modular
   end Initialize_Buffer;

   procedure Initialize (Buf : in out Buffer) renames Initialize_Buffer;

   procedure Insert (Buf : in out Buffer; C : Character) is
   begin
      --  The whole point of this package is that we don't need to adjust the
      --  Markers here! Markers in To_Markers are to the left of the newly
      --  inserted character, so their Position is correct. Valid Markers in
      --  From_Markers (in From_Markers(From_Markers_First..From_Markers'Last))
      --  have Positions relative to From, so also don't need to be adjusted. A
      --  Marker's Position only needs to be adjusted when 'point' is moved
      --  forward past the Marker.
      --
      --  There is one minor exception: A marker that is at 'point' needs to be
      --  adjusted, but only if 'point' is not at the end, as follows:

      if
        not Is_Empty (Buf.To_Markers)
        and then Buf.From_First <= Last_Index (Buf.From)
      then
         declare
            Mark : constant Marker := Last_Element (Buf.To_Markers);

         begin
            if At_Point (Buf, Mark) then
               declare
                  Rec : Marker_Rec renames Buf.Markers (Mx (Buf, Mark));

               begin
                  Rec.Position := Rec.Position + 1;
               end;
            end if;
         end;
      end if;

      Append (Buf.To, C);
      if C = NL then
         Buf.Cur_Column := 1;

      else
         Buf.Cur_Column := Buf.Cur_Column + 1;
      end if;
   end Insert;

   procedure Insert (Buf : in out Buffer; S : String) is
   begin
      for C of S loop
         Insert (Buf, C);
      end loop;
   end Insert;

   procedure Insert_Keeping_Mark
     (Buf  : in out Buffer;
      Mark : Marker;
      C    : Character)
   is
   begin
      pragma Assert (False); -- ???Not used

      Append (Buf.To, C);
      if C = NL then
         Buf.Cur_Column := 1;

      else
         Buf.Cur_Column := Buf.Cur_Column + 1;
      end if;

      pragma Assert (Char_At (Buf, Mark) = C);
   end Insert_Keeping_Mark;

   function Last_Position (Buf : Buffer) return Natural is
   begin
      return
        Last_Index (Buf.To) + (Last_Index (Buf.From) - Buf.From_First + 1);
   end Last_Position;

   function Lookahead
     (Buf : Buffer;
      Offset : Positive := 1)
     return Character
   is
   begin
      if Buf.From_First + Offset <= Last_Index (Buf.From) then
         return Buf.From (Buf.From_First + Offset);

      else
         return Latin_1.NUL;
      end if;
   end Lookahead;

   function Lookback (Buf : Buffer; Offset : Positive := 1) return Character is
   begin
      if Last_Index (Buf.To) - Offset + 1 >= 1 then
         return Buf.To (Last_Index (Buf.To) - Offset + 1);

      else
         return Latin_1.NUL;
      end if;
   end Lookback;

   function Mark (Buf : in out Buffer; Name : Character) return Marker is
   begin
      if Is_Empty (Buf.To_Markers) then
         Append_New_Marker (Buf, Name); -- Create a new one

      --  Avoid creating a new Marker if we already have one at 'point'

      elsif not At_Point (Buf, Last_Element (Buf.To_Markers)) then
         Append_New_Marker (Buf, Name); -- Create a new one
      end if;

      return Last_Element (Buf.To_Markers);
   end Mark;

   function Mark_Previous
     (Buf  : in out Buffer;
      Name : Character)
     return Marker
   is
      L : constant Marker_Index'Base := Last_Index (Buf.To_Markers);
      Insertion_Point : Marker_Index;
      M               : Marker;

   begin
      if
        L >= 1
        and then Position (Buf, Buf.To_Markers (L)) = Last_Index (Buf.To)
      then
         return Buf.To_Markers (L);

      elsif
        L >= 2
        and then Position (Buf, Buf.To_Markers (L - 1)) = Last_Index (Buf.To)
      then
         return Buf.To_Markers (L - 1);

      elsif At_Point (Buf, Buf.To_Markers (L)) then
         Insertion_Point := L;

      else
         Insertion_Point := L + 1;
      end if;

      Append
        (Buf.Markers,
         Marker_Rec'
           (Position => Last_Index (Buf.To),
           Flag      => Buf.To_Flag,
           Name      => Name));
      M := Marker (Last_Index (Buf.Markers)) + Buf.Unique_Id;
      Insert (Buf.To_Markers, Before => Insertion_Point, New_Item => M);

      return M;
   end Mark_Previous;

   function Mx (Buf : Buffer; Mark : Marker) return Marker_Rec_Index is
      Result : constant Marker := Mark - Buf.Unique_Id;

   begin
      if Result not in 1 .. Marker (Last_Index (Buf.Markers)) then
         raise Program_Error
           with "Marker for wrong Buffer: " &
           Image (Modular (Mark)) &
           '-' &
           Image (Modular (Buf.Unique_Id));
      end if;
      return Marker_Rec_Index (Result);
   end Mx;

   function Name (Buf : Buffer; Mark : Marker) return Character is
   begin
      return Buf.Markers (Mx (Buf, Mark)).Name;
   end Name;

   function Point (Buf : Buffer) return Positive is
   begin
      return Last_Index (Buf.To) + 1;
   end Point;

   function Position (Buf : Buffer; Mark : Marker) return Positive is
      Rec : constant Marker_Rec := Buf.Markers (Mx (Buf, Mark));

   begin
      if Rec.Flag = Buf.To_Flag then
         return Rec.Position;

      else
         return Last_Index (Buf.To) - (Buf.From_First - 1) + Rec.Position;
      end if;
   end Position;

   procedure Replace_Cur (Buf : in out Buffer; C : Character) is
   begin
      pragma Assert (not At_End (Buf));
      Buf.From (Buf.From_First) := C;
      --  No need to adjust Buf.Cur_Column
   end Replace_Cur;

   procedure Replace_Previous (Buf : in out Buffer; C : Character) is
   begin
      Buf.To (Last_Index (Buf.To)) := C;
      if C = NL then
         Buf.Cur_Column := 1;
      end if;
   end Replace_Previous;

   procedure Reset (Buf : in out Buffer) is
   begin
      pragma Debug (Validate (Buf, "Reset"));

      pragma Assert (Buf.From_First = Last_Index (Buf.From) + 1);
      if Buf.From_Markers_First = Last_Index (Buf.From_Markers) then
         Move_Forward (Buf);
      end if;
      pragma Assert
        (Buf.From_Markers_First = Last_Index (Buf.From_Markers) + 1);

      Move (Target => Buf.From, Source => Buf.To);
      pragma Assert (Is_Empty (Buf.To));

      Move (Target => Buf.From_Markers, Source => Buf.To_Markers);
      pragma Assert (Is_Empty (Buf.To_Markers));

      Buf.From_First         := 1;
      Buf.From_Markers_First := 1;
      Buf.To_Flag            := not Buf.To_Flag;
      Buf.Cur_Column         := 1;

      --  If there is a Marker pointing to the first character, we need to move
      --  it to To to preserve the invariant.

      Maybe_Move_Marker (Buf);
   end Reset;

   function Slice (Buf : Buffer; First, Last : Marker) return String is
      pragma Assert (Position (Buf, First) <= Position (Buf, Last));
      F : constant Marker_Rec := Buf.Markers (Mx (Buf, First));
      L : constant Marker_Rec := Buf.Markers (Mx (Buf, Last));
      To_F, To_L, From_F, From_l : Natural;

   begin
      if F.Flag = Buf.To_Flag then
         To_F   := F.Position;
         From_F := 1;

      else
         To_F   := Last_Index (Buf.To) + 1;
         From_F := F.Position;
      end if;
      if L.Flag = Buf.To_Flag then
         To_L   := L.Position - 1;
         From_l := 0;

      else
         To_L   := Last_Index (Buf.To);
         From_l := L.Position - 1;
      end if;

      declare
         Result : constant String :=
           Slice (Buf.To, To_F, To_L) & Slice (Buf.From, From_F, From_l);
         pragma Assert (Result'First = 1); -- Ensured by Slice

      begin
         if False then -- Too slow, but we keep it for documentation
            pragma Assert
              (Result =
               To_String (Buf)
                 (Position (Buf, First) .. Position (Buf, Last) - 1));
         end if;
         return Result;
      end;
   end Slice;

   function Slice
     (Buf   : Buffer;
      First : Positive;
      Last  : Natural)
     return String
   is
   begin
      return Result : String (1 .. Last - First + 1) do
         for J in Result'Range loop
            Result (J) := Char_At (Buf, First + J - 1);
         end loop;
         if False then -- Slow
            pragma Assert (Result = To_String (Buf) (First .. Last));
         end if;
      end return;
   end Slice;

   function String_To_Buffer (S : String) return Buffer is
   begin
      return Buf : Buffer do
         Insert (Buf, S);
         Reset (Buf);
      end return;
   end String_To_Buffer;

   function To_Debug_String (Buf : Buffer) return String is
      S     : constant String       := To_String (Buf);
      Marks : constant Marker_Array :=
        To_Array (Buf.To_Markers) &
        Slice
          (Buf.From_Markers,
           Buf.From_Markers_First,
           Last_Index (Buf.From_Markers));
      pragma Assert (Marks'First = 1);
      M_Index : Marker_Index := 1;
      Result : Char_Vector;

   --  Start of processing for To_Debug_String

   begin
      for X in S'Range loop
         if M_Index <= Marks'Last then
            pragma Assert (X <= Position (Buf, Marks (M_Index)));
            if X = Position (Buf, Marks (M_Index)) then
               Append (Result, Name (Buf, Marks (M_Index)));
               M_Index := M_Index + 1;
            end if;

            while M_Index <= Marks'Last
            and then X = Position (Buf, Marks (M_Index))
            loop
               Append (Result, "[duplicate]");
               Append (Result, Name (Buf, Marks (M_Index)));
               M_Index := M_Index + 1;
            end loop;
         end if;
         Append (Result, S (X));
      end loop;

      --  Check for one last marker after the last character

      if M_Index = Marks'Last then
         Append (Result, Name (Buf, Marks (M_Index)));
         M_Index := M_Index + 1;
      end if;
      pragma Assert (M_Index = Marks'Last + 1);

      return To_Array (Result);
   end To_Debug_String;

   function To_String (Buf : Buffer) return String is
   begin
      return
        Result : constant String := To_Array (Buf.To) & Valid_From (Buf)
      do
         pragma Assert (Result'First = 1);
      end return;
   end To_String;

   function Valid_From (Buf : Buffer) return String is
   begin
      return Slice (Buf.From, Buf.From_First, Last_Index (Buf.From));
   end Valid_From;

   Validations : Natural := 0;

   procedure Validate (Buf : Buffer; Message : String) is

      procedure Fail (Msg : String);

      procedure Fail (Msg : String) is
      begin
         Dbg_Out.Output_Enabled := True;
         Dbg_Out.Put ("Validate: \1: failure: \2\n", Message, Msg);
         Dump_Buffer (Buf);
         raise Program_Error;
      end Fail;

      function Marker_Less (X, Y : Marker) return Boolean;

      function Marker_Less (X, Y : Marker) return Boolean is
      begin
         return Position (Buf, X) < Position (Buf, Y);
      end Marker_Less;

      package Sorting is new Marker_Vectors.Generic_Sorting
        ("<" => Marker_Less);

      Marks : constant Marker_Array :=
        To_Array (Buf.To_Markers) &
        Slice
          (Buf.From_Markers,
           Buf.From_Markers_First,
           Last_Index (Buf.From_Markers));
      pragma Assert (Marks'First = 1);

   --  Start of processing for Validate

   begin
      if False then
         Dbg_Out.Output_Enabled := True;
         Dbg_Out.Put ("\nValidate: \1\n", Message);
         Dump_Buffer (Buf);
      end if;

      if not Sorting.Is_Sorted (Buf.To_Markers) then
         Fail ("To_Markers not sorted");
      end if;
      if not Sorting.Is_Sorted (Buf.From_Markers) then
         Fail ("From_Markers not sorted");
      end if;

      for M in 1 .. Last_Index (Buf.To_Markers) loop
         declare
            Mark : constant Marker     := Buf.To_Markers (M);
            Rec  : constant Marker_Rec := Buf.Markers (Mx (Buf, Mark));

         begin
            if Rec.Flag /= Buf.To_Flag then
               Fail ("Bad To_Marker flag");
            end if;
            pragma Warnings (Off);
            --  "lower bound check only fails if it is invalid"
            if Rec.Position not in 1 .. Last_Index (Buf.To) + 1 then
               pragma Warnings (On);
               Fail ("Bad To_Marker position");
            end if;
         end;
      end loop;

      for M in Buf.From_Markers_First .. Last_Index (Buf.From_Markers) loop
         declare
            Mark : constant Marker     := Buf.From_Markers (M);
            Rec  : constant Marker_Rec := Buf.Markers (Mx (Buf, Mark));

         begin
            if Rec.Flag /= not Buf.To_Flag then
               Fail ("Bad From_Marker flag");
            end if;
            if
              Rec.Position not in
                Buf.From_First + 1 .. Last_Index (Buf.From) + 1
            then
               Fail ("Bad From_Marker position");
            end if;
         end;
      end loop;

      for J in 2 .. Marks'Last loop
         if Position (Buf, Marks (J - 1)) = Position (Buf, Marks (J)) then
            Fail ("Duplicate marks");
         end if;
      end loop;

      Validations := Validations + 1;
   end Validate;

end Ada_Trees.Buffers;
