------------------------------------------------------------------------------
--                                                                          --
--                            GNAT2XML COMPONENTS                           --
--                                                                          --
--                    G N A T 2 X M L . A D A _ T R E E S                   --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                  Copyright (C) 2012-2013, AdaCore, Inc.                  --
--                                                                          --
-- Gnat2xml is free software; you can redistribute it and/or modify it      --
-- under terms of the  GNU General Public License  as published by the Free --
-- Software Foundation;  either version 2,  or  (at your option)  any later --
-- version. Gnat2xml is distributed  in the hope  that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of MER-      --
-- CHANTABILITY or  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General  --
-- Public License for more details. You should have received a copy of the  --
-- GNU General Public License distributed with GNAT; see file COPYING. If   --
-- not, write to the Free Software Foundation, 59 Temple Place Suite 330,   --
-- Boston, MA 02111-1307, USA.                                              --
-- The gnat2xml tool was derived from the Avatox sources.                   --
------------------------------------------------------------------------------

pragma Ada_2012;

with Ada.Exceptions;

with Snames;

with ASIS_UL.Generic_Formatted_Output;

separate (Ada_Trees.Formatting)

procedure Tree_To_Ada
  (Tree : Ada_Tree; Source_Text : String; Options : Formatting_Options)
is
   --  Miscellaneous useful Name_Ids:

   Name_Semicolon : constant Name_Id := Name_Find (";");
   Name_L_Paren   : constant Name_Id := Name_Find ("(");
   Name_R_Paren   : constant Name_Id := Name_Find (")");
   Name_Colon     : constant Name_Id := Name_Find (":");
   Name_Bang      : constant Name_Id := Name_Find ("!");
   Name_Bar       : constant Name_Id := Name_Find ("|");
   Name_Dot       : constant Name_Id := Name_Find (".");

   Name_And_Then : constant Name_Id := Name_Find ("and then");
   Name_Or_Else  : constant Name_Id := Name_Find ("or else");

   Name_Q_And : constant Name_Id := Name_Find ("""and""");
   Name_Q_Or  : constant Name_Id := Name_Find ("""or""");
   Name_Q_Xor : constant Name_Id := Name_Find ("""xor""");
   Name_Q_Mod : constant Name_Id := Name_Find ("""mod""");
   Name_Q_Rem : constant Name_Id := Name_Find ("""rem""");
   Name_Q_Abs : constant Name_Id := Name_Find ("""abs""");
   Name_Q_Not : constant Name_Id := Name_Find ("""not""");

   Op_Sym_Table : constant array (Positive range <>) of Name_Id :=
     (Name_Q_And,
      Name_Q_Or,
      Name_Q_Xor,
      Name_Q_Mod,
      Name_Q_Rem,
      Name_Q_Abs,
      Name_Q_Not);

   function Is_Op_Sym_With_Letters
     (N                           : Name_Id) return Boolean is
     (for some Op of Op_Sym_Table => N = Op);
   --  True if N looks like a string literal that can be used as an operator
   --  symbol containing letters, so case might matter. N should be in all
   --  lower case.

   procedure Put_To_Buffer (C : Character);
   --  Append C to Buffer

   Check_Whitespace : Boolean := False;
   --  For debugging. True during the Subtree_To_Ada phase. Check for two
   --  blanks in a row, or blanks at end of line. ???Check the latter always?

   procedure Put_To_Buffer (C : Character) is
      use Scanner;

   begin
      if Check_Whitespace then
         if C in ' ' | NL then
            pragma Assert (Lookback (Out_Buf) /= ' ');
         end if;
      end if;

      Insert (Out_Buf, C);

      if Debug_Mode then
         if C = NL then
            Text_IO.Put (Text_IO.Standard_Output, "<NL>");
            Text_IO.New_Line (Text_IO.Standard_Output);

         else
            Text_IO.Put (Text_IO.Standard_Output, C);
         end if;
      end if;
   end Put_To_Buffer;

   package Buffered_Output is new ASIS_UL.Generic_Formatted_Output
     (Basic_Put_Char => Put_To_Buffer);

   procedure Indent (Amount : Integer);

   procedure Indent (Amount : Integer) is
      Line_Breaks : Line_Break_Vector renames All_Line_Breaks;

   begin
      Cur_Indentation := Cur_Indentation + Amount;

      pragma Assert
        (abs (Amount) in 0 | 1 |
           Options.PP_Indentation | Options.PP_Cont_Line_Indentation);
      if abs (Amount) = Options.PP_Indentation then -- ???
         pragma Assert (Point (Out_Buf) = Last_Position (Out_Buf) + 1);
         if
           Last_Position (Out_Buf) =
           Position (Out_Buf, Line_Breaks (Last (Line_Breaks)).Mark)
         then
--  pragma Assert (At_Point (Out_Buf, Line_Breaks (Last (Line_Breaks)).Mark));
            Line_Breaks (Last (Line_Breaks)).Indentation := Cur_Indentation;
         end if;
      end if;
   end Indent;

   procedure Append_Line_Break
     (Hard     : Boolean;
      Level    : Nesting_Level;
      Kind     : Ada_Tree_Kind;
      Template : Name_Id);

   procedure Append_Temp_Line_Break;

   procedure Append_Line_Break
     (Hard     : Boolean;
      Level    : Nesting_Level;
      Kind     : Ada_Tree_Kind;
      Template : Name_Id)
   is

      Line_Breaks : Line_Break_Vector renames All_Line_Breaks;

   begin
      Append (Line_Breaks, (Mark => Mark (Out_Buf, Name => (if Hard then
                   '$'
                else
                   '@')),
         Hard              => Hard,
         Enabled           => Hard,
         Level             => Level,
         Indentation       => Cur_Indentation,
         Extra_Indentation => 0,
         Length            => <>,
         Kind              => Kind,
         Template          => Template));

                  --  A hard line break gets NL

      if Hard then
         Buffered_Output.Put_Char (NL);
      end if;
   end Append_Line_Break;

   procedure Append_Temp_Line_Break is
      M : Marker;

   begin
      Insert (Out_Buf, NL);
      M := Mark_Previous (Out_Buf, Name => '-');

      if False then -- Too slow, but we keep it for documentation
         for L of All_Line_Breaks loop
            pragma Assert (M /= L.Mark);
         end loop;
      end if;

      Append
        (Temp_Line_Breaks,
         (Mark             => M,
         Hard              => True,
         Enabled           => True,
         Level             => 0,
         Indentation       => Cur_Indentation,
         Extra_Indentation => 0,
         Length            => <>,
         Kind              => Not_An_Element,
         Template          => Name_Find ("Insert_Comments_And_Blank_Lines")));
      pragma Assert (Char_At (Out_Buf, M) = NL);
   end Append_Temp_Line_Break;

   --  ???Move following procedures

   function Max_Nesting_Increment (Temp : Ada_Template) return Nesting_Level;

   function Max_Nesting_Increment (Temp : Ada_Template) return Nesting_Level is
      J : Positive := Temp'First;
      C : Character;

   begin
      return Result : Nesting_Level := 0 do
         while J <= Temp'Last loop
            C := Temp (J);

            case C is
               when '[' | ']' | '@' =>
                  declare
                     Digit     : Character;
                     Increment : Nesting_Level;

                  begin
                     if J < Temp'Last and then Temp (J + 1) in '0' .. '9' then
                        J         := J + 1;
                        Digit     := Temp (J);
                        Increment := Nesting_Level (Char_To_Digit (Digit));

                     else
                        Increment := 0;
                     end if;

                     Result := Nesting_Level'Max (Result, Increment);
                  end;

               when others =>
                  null;
            end case;

            J := J + 1;
         end loop;
      end return;
   end Max_Nesting_Increment;

   Inner_Loop_Count : Natural := 0;

   procedure Split_Lines;

   procedure Split_Lines is
      Line_Breaks : Line_Break_Vector renames All_Line_Breaks;

      procedure Remove_Duplicates;
      --  Remove soft line breaks that have the same Mark as other line
      --  break(s). This is necessary because we don't want line breaks to form
      --  blank lines.

      function Line_Length (F, L : Line_Break_Index) return Natural;
      --  F and L are the first and last index forming a line; returns the
      --  length of the line, not counting new-lines. F and L must be enabled.

      function Worthwhile_Line_Break (X : Line_Break_Index) return Boolean;
      --  Called for the first so-far-disabled line break on a line. Returning
      --  False means don't bother enabling it.

      function Line_Length (F, L : Line_Break_Index) return Natural is
         First : constant Line_Break := Line_Breaks (F);
         Last  : constant Line_Break := Line_Breaks (L);
         pragma Assert (First.Enabled and then Last.Enabled);
         NL_Count : constant Natural := (if First.Hard then
               1
            else
               0);
         Without_Indent : constant Natural :=
           Position (Out_Buf, Last.Mark) -
           Position (Out_Buf, First.Mark) -
           NL_Count;
      --  The length without the indentation is just the difference between
      --  the two marks, except that if the first one is hard, we don't
      --  count the NL character. If it's soft, there is no NL character
      --  yet.

      begin
         --  If the line is blank, we ignore the indentation; we won't be
         --  putting blanks in the output. Otherwise, the length is the
         --  indentation plus the length without the indentation as calculated
         --  above.

         if Without_Indent = 0 then
            return 0;

         else
            return First.Indentation + Without_Indent;
         end if;
      end Line_Length;

      procedure Remove_Duplicates is
         Temp : Line_Break_Vector;

      begin
         Append (Temp, Line_Breaks (1));

         for X in 2 .. Last_Index (Line_Breaks) loop
            if
              Line_Breaks (X).Enabled or else not Is_Empty_Line (X - 1, X)
            then
               Append (Temp, Line_Breaks (X));

            else
               pragma Assert (not Line_Breaks (X).Hard);
            end if;
         end loop;
         Move (Target => Line_Breaks, Source => Temp);
      end Remove_Duplicates;

      function Worthwhile_Line_Break (X : Line_Break_Index) return Boolean is
         This : constant Positive := Position (Out_Buf, Line_Breaks (X).Mark);
         Prev : Positive := Position (Out_Buf, Line_Breaks (X - 1).Mark);

      begin
         if Line_Breaks (X - 1).Hard then
            Prev := Prev + 1; -- skip NL
         end if;

         --  If we have something like:
         --     P (...
         --  there's no point in turning it into:
         --     P
         --       (...
         --  assuming PP_Cont_Line_Indentation = 2, because it doesn't shorten
         --  any lines. If the procedure name is slightly longer than "P":
         --     Proc (...
         --  there's _probably_ no point in turning it into:
         --     Proc
         --       (...
         --  because it only saves 3 characters, so we will probably have to
         --  split up the "..." parameters anyway. ???Not sure about this
         --  heuristic.
         --
         --  ???The "if " test is to avoid '(style) misplaced "then"'.

         declare
            Line_Start : constant String := Slice (Out_Buf, Prev, This);

         begin
            if To_Lower (Line_Start) = "if " then
               return True;
            end if;
         end;
         if This - Prev <= Options.PP_Cont_Line_Indentation then
            return False;
         end if;
         return True;
      end Worthwhile_Line_Break;

      F : Line_Break_Index := 1;
      L   : Line_Break_Index;
      Len : Natural;

      Level       : Nesting_Level;
      More_Levels : Boolean;

   --  Start of processing for Split_Lines

   begin
      Remove_Duplicates;
      if False then
         --  ???For debugging, always split at optional newlines
         for Line_Index in 1 .. Last_Index (Line_Breaks) loop
            Line_Breaks (Line_Index).Enabled := True;
         end loop;
         return;
      end if;

      while F /= Last_Index (Line_Breaks) loop
         Level       := 1;
         More_Levels := True;

         loop -- through levels
            L   := Next_Enabled (F);
            Len := Line_Length (F, L);
            exit when Len <= Options.Max_Line_Length; -- short enough
            exit when not More_Levels; -- no more line breaks to enable

            More_Levels := False;

            for X in F + 1 .. L - 1 loop
               if Line_Breaks (X).Level > Level then
                  More_Levels := True;

               elsif Line_Breaks (X).Level = Level then
                  Inner_Loop_Count := Inner_Loop_Count + 1;

                  --  Don't enable the first one, unless it's "worthwhile"
                  --  according to the heuristic.
                  if X = F + 1 and then not Worthwhile_Line_Break (X) then
                     null;

                  --  We don't want soft line breaks to form blank lines, so
                  --  don't enable this one if the previous one is already
                  --  enabled.

                  elsif
                    Line_Breaks (X - 1).Enabled
                    and then Is_Empty_Line (X - 1, X)
                  then
                     raise Program_Error;

                  else
                     pragma Assert (if Line_Breaks (X - 1).Enabled then
                           Line_Breaks (X - 1).Mark /= Line_Breaks (X).Mark);
                     Line_Breaks (X).Enabled := True;
                  end if;
               end if;
            end loop;

            Level := Level + 1;
         end loop; -- through levels

         Line_Breaks (F).Length := Len;
         F                      := L;
      end loop; -- through line breaks

      Line_Breaks (F).Length := 0; -- last line

      --  Assert that the line Length has been set if and only if the line
      --  break is enabled.

      for X in 1 .. Last_Index (Line_Breaks) loop
         declare
            Break : constant Line_Break := Line_Breaks (X);

         begin
            if X = Last_Index (Line_Breaks) then
               pragma Assert (Break.Enabled and then Break.Length = 0);

            elsif Break.Enabled then
               pragma Assert
                 (Break.Length = Line_Length (X, Next_Enabled (X)));
               pragma Assert
                 (Break.Mark /= Line_Breaks (Next_Enabled (X)).Mark);

            else
               pragma Assert (Break.Length = Natural'Last);
            end if;
         end;
      end loop;
   end Split_Lines;

   procedure Insert_NLs_And_Indentation;

   procedure Insert_NLs_And_Indentation is
      --  We loop through Out_Buf, and for each character, take care of the
      --  Line_Break at that character, if any. The Line_Breaks are in
      --  Enabled_Line_Breaks. Enabled_Line_Breaks cannot have duplicates (two
      --  elements at the same Mark), because hard line breaks take up space in
      --  Out_Buf (there is an NL), and we never enable two soft line breaks in
      --  a row.

      At_Line_Start : Boolean := True;
      Indentation   : Natural := 0;

      Cur_Line : Line_Break_Index := 1;
      Line_Breaks : Line_Break_Vector renames Enabled_Line_Breaks;

   begin
      Collect_Enabled_Line_Breaks;

      Char_Loop : loop
         pragma Assert
           (Position (Out_Buf, Last_Element (Tabs).Mark) =
            Last_Position (Out_Buf) + 1);

         pragma Assert
           (Point (Out_Buf) <=
            Position (Out_Buf, Line_Breaks (Cur_Line).Mark));
         if At_Point (Out_Buf, Line_Breaks (Cur_Line).Mark) then
            Dbg_Out.Put ("\n");
         end if;

         --  Even though Enabled_Line_Breaks cannot have duplicates, we still
         --  need 'while' (not 'if'), because in one case we Move_Forward
         --  below.

         while At_Point (Out_Buf, Line_Breaks (Cur_Line).Mark) loop
            pragma Assert
              (Point (Out_Buf) =
               Position (Out_Buf, Line_Breaks (Cur_Line).Mark));
            Dbg_Out.Put ("Point = \1, break = ", Image (Point (Out_Buf)));
            Dump_Marker (Out_Buf, Line_Breaks (Cur_Line).Mark);

            At_Line_Start := True;

            --  A hard line break already has NL; for a soft one, we need to
            --  add NL

            if Line_Breaks (Cur_Line).Hard then
               Dbg_Out.Put
                 ("\1: hard line break\n",
                  Image (Integer (Cur_Line)));
               pragma Assert (Cur (Out_Buf) = NL);
               Move_Forward (Out_Buf);

            else
               --  A soft line break can be preceded or followed by a blank,
               --  but never both, and never more than one. If there is a
               --  blank, we replace it with NL, otherwise we insert NL.

               if Lookback (Out_Buf) = ' ' then
                  Dbg_Out.Put
                    ("\1: soft line break Replace_Previous\n",
                     Image (Integer (Cur_Line)));
                  pragma Assert (Cur (Out_Buf) /= ' ');
                  Replace_Previous (Out_Buf, NL);
                  pragma Assert
                    (not At_Point (Out_Buf, Line_Breaks (Cur_Line + 1).Mark));

               elsif Cur (Out_Buf) = ' ' then
                  Dbg_Out.Put
                    ("\1: soft line break Replace_Cur\n",
                     Image (Integer (Cur_Line)));
                  Replace_Cur (Out_Buf, NL);
                  pragma Assert
                    (not At_Point (Out_Buf, Line_Breaks (Cur_Line + 1).Mark));
                  Move_Forward (Out_Buf);

               else
                  Dbg_Out.Put
                    ("\1: soft line break insert\n",
                     Image (Integer (Cur_Line)));
                  Insert (Out_Buf, NL);
                  pragma Assert
                    (not At_Point (Out_Buf, Line_Breaks (Cur_Line + 1).Mark));
               end if;
            end if;
            Indentation := Line_Breaks (Cur_Line).Indentation;

            pragma Assert
              (At_End (Out_Buf) = (Cur_Line = Last_Index (Line_Breaks)));
            exit Char_Loop when Cur_Line = Last_Index (Line_Breaks);

            Cur_Line := Cur_Line + 1;
            Dbg_Out.Put
              ("    point = \1, next break = ",
               Image (Point (Out_Buf)));
            Dump_Marker (Out_Buf, Line_Breaks (Cur_Line).Mark);
            Dbg_Out.Put ("\n");
            pragma Assert
              (Point (Out_Buf) <=
               Position (Out_Buf, Line_Breaks (Cur_Line).Mark));
         end loop; -- through Line_Breaks table

         --  We can't be At_End, because we would have done "exit Char_Loop"
         --  above.

         pragma Assert (not At_End (Out_Buf));
         pragma Assert (Cur (Out_Buf) not in NL | ASCII.NUL);

         if At_Line_Start then
            for J in 1 .. Indentation loop
               Insert (Out_Buf, ' ');
            end loop;
            At_Line_Start := False;
         end if;

         Move_Forward (Out_Buf);
      end loop Char_Loop;

      pragma Assert (At_End (Out_Buf));
      pragma Assert (Cur_Line = Last_Index (Line_Breaks));
   end Insert_NLs_And_Indentation;

   procedure Write_Buffer (Buf : in out Buffer);

   procedure Write_Buffer (Buf : in out Buffer) is
      pragma Assert (Point (Buf) = 1);
      pragma Assert (Char_At (Buf, Last_Position (Buf)) = NL);
      First_Time : Boolean := True;

   begin
      if False then -- ???Messes up the diff's.
         Formatted_Output.Put
           ("--  ???Inner_Loop_Count = \1\n",
            Image (Inner_Loop_Count));
      end if;

      Outer_Loop : loop
         declare
            NL_Count : Natural := 0;

         begin
            while Cur (Buf) = NL loop
               if Point (Buf) = Last_Position (Buf) then
                  Formatted_Output.Put_Char (NL);
                  --  Skip extra NL's at end of file
                  exit Outer_Loop;
               end if;
               NL_Count := NL_Count + 1;
               Move_Forward (Buf);
            end loop;

            if First_Time then
               --  Skip all NL's at start of file
               First_Time := False;
               NL_Count   := 0;
            end if;

            if NL_Count >= 1 then
               Formatted_Output.Put_Char (NL);
               if NL_Count >= 2 then
                  Formatted_Output.Put_Char (NL);
               end if;
            end if;
         end;

         pragma Assert (Cur (Buf) /= NL);
         Formatted_Output.Put_Char (Cur (Buf));
         Move_Forward (Buf);
      end loop Outer_Loop;
   end Write_Buffer;

   procedure Subtree_To_Ada
     (Tree            : Ada_Tree;
      Cur_Level       : Nesting_Level;
      Index_In_Parent : Query_Index);
   --  We recursively walk the tree, and for most nodes, take the template
   --  from Template_Table, and pass it to Interpret_Template. Some nodes
   --  need special casing, and bypass the Template_Table. Subtree_To_Ada is
   --  directly recursive, and also mutually recursive with
   --  Interpret_Template.

   package Tree_Stacks is new Ada.Containers.Vectors
     (Index_Type   => Query_Index,
      Element_Type => Ada_Tree);
   use Tree_Stacks;

   Tree_Stack : Tree_Stacks.Vector;
   --  Stack of trees that we're in the process of traversing. Pushed and
   --  popped at the beginning and end of Subtree_To_Ada.

   function Ancestor_Tree
     (N : Query_Count) return Ada_Tree is
     (Tree_Stack (Last_Index (Tree_Stack) - N));
   --  Returns the N'th ancestor of the current tree.
   --  Ancestor (0) is the current tree,
   --  Ancestor (1) is the parent of the current tree,
   --  Ancestor (2) is the grandparent of the current tree,
   --  and so on.

   function Parent_Tree return Ada_Tree is (Ancestor_Tree (1));

   L_Square_Seen : Boolean := False;
--   Square_Nesting : Natural := 0; -- ???Redundant with Cur_Level?

   Implicit_Null_Statement_Kludge : Boolean := False;
   --  See the comments about labels under "when A_Null_Statement =>" below for
   --  an explanation of this.

   procedure Subtree_To_Ada
     (Tree            : Ada_Tree;
      Cur_Level       : Nesting_Level;
      Index_In_Parent : Query_Index)
   is
      pragma Unreferenced (Index_In_Parent); -- ???Needed?

      Line_Breaks : Line_Break_Vector renames All_Line_Breaks;

      procedure Subtrees_To_Ada
        (Tree               : Ada_Tree;
         Pre, Between, Post : Ada_Template);

      procedure Interpret_Template
        (T : Ada_Template := Template_Table (Tree.Kind).all;
         Subtrees : Ada_Tree_Array := Tree.Subtrees;
         Cur_Level : Nesting_Level := Subtree_To_Ada.Cur_Level;
         Kind : Ada_Tree_Kind := Tree.Kind);
      --  Interpret the template, printing literal characters, and recursively
      --  calling Subtree_To_Ada when the template calls for a subnode. Kind is
      --  for debugging.

      procedure Prefix_Notation_Call (Label_Names, Callee, Actuals : Ada_Tree);
      --  This is called for A_Function_Call and A_Procedure_Call_Statement
      --  when the Is_Prefix_Notation subtree is True. Prefix notation calls
      --  have special visibility rules, so we don't want to turn X.F(Y)
      --  into F(X, Y). Label_Names is always empty for function calls.

      procedure Append_Tab
        (T             : Ada_Template;
         Token_Text    : Name_Id;
         Index_In_Line : Tab_Index_In_Line);
      --  Append a Tab_Rec onto Tabs. If Token is Name_Empty, get the token
      --  from the template T.

      procedure Append_Tab
        (T             : Ada_Template;
         Token_Text    : Name_Id;
         Index_In_Line : Tab_Index_In_Line)
      is

         Tok  : Scanner.Token;
         Text : Name_Id;
         pragma Unreferenced (Text);

      begin
         if Token_Text = Name_Empty then
            Tok  := Scanner.Get_Token (String (T));
            Text := Tok.Normalized;
            pragma Assert (Tok.Sloc.First = 1);

         else
            Text := Token_Text;
         end if;

         pragma Assert
           (Point (Out_Buf) =
            Last_Position (Out_Buf) + 1); -- ???Do we need Last_Position?
         Append
           (Tabs,
            (Token        => Token_Text, -- Should be Text???
            Mark          => Mark (Out_Buf, '^'),
            Index_In_Line => Index_In_Line,
            Col           => <>,
            Num_Blanks    => <>));
         pragma Assert
           (Position (Out_Buf, Last_Element (Tabs).Mark) =
            Last_Position (Out_Buf) + 1);
      end Append_Tab;

      procedure Subtrees_To_Ada
        (Tree               : Ada_Tree;
         Pre, Between, Post : Ada_Template)
      is
         pragma Assert (Tree.Kind in Flat_List_Kinds);
         First_Time : Boolean := True;
         Prev : Ada_Tree;

      begin
         if Tree.Subtree_Count = 0 then
            return;
         end if;

         Interpret_Template (Pre, Subtrees => Empty_Tree_Array);

         for Index in 1 .. Tree.Subtree_Count loop
            declare
               Subtree : constant Ada_Tree := Tree.Subtrees (Index);

            begin
               if First_Time then
                  First_Time := False;

               else
                  --  Special case for use_package_clauses: We want to print
                  --  "with A.B; use A.B;" on one line.

                  pragma Warnings (Off, "*""Prev"" may be null"); -- May not!
                  if
                    Subtree.Kind in A_Use_Package_Clause
                    and then Subtree.Subtrees (1).Subtree_Count = 1
                    and then Prev.Kind in A_With_Clause
                    and then Prev.Subtrees (3).Subtree_Count = 1
                    and then
                      Ref (Subtree.Subtrees (1).Subtrees (1)) =
                      Ref (Prev.Subtrees (3).Subtrees (1))
                  then
                     pragma Warnings (On, "*""Prev"" may be null");

                     pragma Assert (Between = ";$");
                     Interpret_Template ("; ", Subtrees => Empty_Tree_Array);

                     Append_Tab ("", Snames.Name_Use, Index_In_Line => 1);

                  else
                     Interpret_Template
                       (Between,
                        Subtrees => Empty_Tree_Array);
                  end if;
               end if;

               Subtree_To_Ada
                 (Subtree,
                  Cur_Level +
                  Max_Nesting_Increment (Pre & Between & Post) +
                  1,
                  Index);
               --  ???Shouldn't this use the entire template?

               Prev := Subtree;
            end;
         end loop;

         Interpret_Template (Post, Subtrees => Empty_Tree_Array);
      end Subtrees_To_Ada;

      procedure Interpret_Template
        (T : Ada_Template := Template_Table (Tree.Kind).all;
         Subtrees : Ada_Tree_Array := Tree.Subtrees;
         Cur_Level : Nesting_Level := Subtree_To_Ada.Cur_Level;
         Kind : Ada_Tree_Kind := Tree.Kind)
      is
         J : Positive := T'First;
         subtype Subtrees_Index is Query_Index range 1 .. Subtrees'Last;
         Used : array (Subtrees_Index) of Boolean := (others => False);
         Cur_Subtree_Index : Query_Count := 0;
         Numeric_Arg : Boolean;
         C           : Character;

         function Debug_Template return Name_Id;

         function Debug_Template return Name_Id is
         begin
            if False then
               return
                 Name_Find ("X" & String (T) & "X    [" & Image (J) & "]");

            else
               return Name_Empty;
            end if;
         end Debug_Template;

         Nesting_Increment : Nesting_Level;

      --  Start of processing for Interpret_Template

      begin
         while J <= T'Last loop
            Numeric_Arg := False;
            C           := T (J);

            case C is
               when '[' =>
                  L_Square_Seen := True;

               when ' ' | '?' =>
                  null;

               when '(' =>
                  if L_Square_Seen then
                     declare
                        Top : constant Line_Break_Index :=
                          Last_Element (Line_Stack);

                     begin
                        pragma Assert
                          (Line_Breaks (Top).Extra_Indentation = 0);
                        Line_Breaks (Top).Extra_Indentation := 1;
                        Indent (1);
                        L_Square_Seen := False;
                     end;
                  end if;

               when others =>
                  L_Square_Seen := False;
            end case;

            case C is
               when '$' =>
                  Append_Line_Break
                    (Hard     => True,
                     Level    => Cur_Level,
                     Kind     => Kind,
                     Template => Debug_Template);

               when '{' =>
                  pragma Assert (T (J - 1) = '$');
                  Indent (Options.PP_Indentation);

               when '}' =>
                  Indent (-Options.PP_Indentation);

               when '[' =>
                  if J < T'Last and then T (J + 1) in '0' .. '9' then
                     J                 := J + 1;
                     Nesting_Increment :=
                       Nesting_Level (Char_To_Digit (T (J)));

                  else
                     Nesting_Increment := 0;
                  end if;
                  Indent (Options.PP_Cont_Line_Indentation);
                  Append_Line_Break
                    (Hard     => False,
                     Level    => Cur_Level + Nesting_Increment,
                     Kind     => Kind,
                     Template => Debug_Template);
                  Append (Line_Stack, Last_Index (Line_Breaks)); -- push
--                  Square_Nesting := Square_Nesting + 1;

               when ']' =>
                  Indent (-Options.PP_Cont_Line_Indentation);
                  Indent
                    (-Line_Breaks (Last_Element (Line_Stack))
                       .Extra_Indentation);
                  Delete_Last (Line_Stack); -- pop
--                  Square_Nesting := Square_Nesting - 1;

               when '@' =>
                  if J < T'Last and then T (J + 1) in '0' .. '9' then
                     J                 := J + 1;
                     Nesting_Increment :=
                       Nesting_Level (Char_To_Digit (T (J)));

                  else
                     Nesting_Increment := 0;
                  end if;
                  Append_Line_Break
                    (Hard     => False,
                     Level    => Cur_Level + Nesting_Increment,
                     Kind     => Kind,
                     Template => Debug_Template);

               when '^' =>
                  declare
                     Index_In_Line : Tab_Index_In_Line;

                  begin
                     if J < T'Last and then T (J + 1) in '0' .. '9' then
                        J             := J + 1;
                        Index_In_Line :=
                          Tab_Index_In_Line (Char_To_Digit (T (J)));

                     else
                        Index_In_Line := 1;
                     end if;
                     Append_Tab
                       (T (J + 1 .. T'Last),
                        Name_Empty,
                        Index_In_Line => Index_In_Line);
                  end;

               when '!' | '?' =>
                  if J < T'Last and then T (J + 1) in '0' .. '9' then
                     Numeric_Arg := True;
                     J           := J + 1;

                  else
                     Cur_Subtree_Index := Cur_Subtree_Index + 1;
                  end if;

                  declare
                     Subtree_Index : Query_Index;

                  begin
                     if Numeric_Arg then
                        Subtree_Index := Query_Index (Char_To_Digit (T (J)));

                     else
                        Subtree_Index := Cur_Subtree_Index;
                     end if;
                     if Subtree_Index not in Subtrees_Index then -- ???
                        raise Program_Error;
                        --  Put ("Subtree_Index = \1, not in \2..\3 <<\4>>\n",
                        --  Image (Subtree_Index),
                        --  Image (Subtrees'First),
                        --  Image (Subtrees'Last),
                        --  "???Image (Tr.Kind)");
                     end if;

                     declare
                        Subtree : constant Ada_Tree :=
                          Subtrees (Subtree_Index);

                     begin
                        Used (Subtree_Index) := True;
                        if C = '!' then
                           Subtree_To_Ada
                             (Subtree,
                              Cur_Level + Max_Nesting_Increment (T) + 1,
                              Subtree_Index);

                        else
                           pragma Assert (C = '?');

                           declare
                              function Scan_To_Tilde return Positive;

                              function Scan_To_Tilde return Positive is
                              begin
                                 loop
                                    J := J + 1;
                                    exit when T (J) = '~';
                                 end loop;
                                 return J - 1;
                              end Scan_To_Tilde;

                              Pre_First : constant Positive := J + 1;
                              Pre_Last  : constant Positive := Scan_To_Tilde;
                              pragma Assert (T (J) = '~');

                              Between_First : constant Positive := J + 1;
                              Between_Last  : constant Positive :=
                                Scan_To_Tilde;
                              pragma Assert (T (J) = '~');

                              Post_First : constant Positive := J + 1;
                              Post_Last  : constant Positive := Scan_To_Tilde;
                              pragma Assert (T (J) = '~');

                           begin
                              Used (Subtree_Index) := True;
                              case Subtree.Kind is -- ???This is kind of kludgy
                                 when Flat_List_Kinds =>
                                    Append (Tree_Stack, Subtree); -- push
                                    Subtrees_To_Ada
                                      (Subtree,
                                       T (Pre_First .. Pre_Last),
                                       T (Between_First .. Between_Last),
                                       T (Post_First .. Post_Last));
                                    Delete_Last (Tree_Stack); -- pop

                                 when Not_An_Element =>
                                    null;

                                 when others =>
                                    Interpret_Template
                                      (T (Pre_First .. Pre_Last),
                                       Subtrees => Empty_Tree_Array);
                                    --  ???
                                    --  if False and then Between /= "" then
                                    --  Put ("\1, \2: ???Between = <<\3>>,
                                    --  " & "T = <<\4>>\n",
                                    --  "???Image (Tr.Kind)",
                                    --  Image (Subtree.Kind),
                                    --  String (Between),
                                    --  String (T));
                                    --  pragma Assert (Between = "");
                                    --  end if;
                                    Subtree_To_Ada
                                      (Subtree,
                                       Cur_Level +
                                       Max_Nesting_Increment (T) +
                                       1,
                                       Subtree_Index);
                                    Interpret_Template
                                      (T (Post_First .. Post_Last),
                                       Subtrees => Empty_Tree_Array);
                              end case;
                           end;
                        end if;
                     end;
                  end;

               when ';' =>
                  if Implicit_Null_Statement_Kludge then
                     Implicit_Null_Statement_Kludge := False;

                  else
                     Buffered_Output.Put_Char (C);
                  end if;

               when others =>
                  Buffered_Output.Put_Char (C);

            end case;

            J := J + 1;
         end loop;

         if Used /= (Subtrees_Index => True) then -- ???
            Buffered_Output.Put ("???Not all used: \1", "???Image (Tr.Kind)");
         end if;
         --  ???pragma Assert (Used = (Subtrees_Index => True));
      end Interpret_Template;

      procedure Prefix_Notation_Call
        (Label_Names, Callee, Actuals : Ada_Tree)
      is

         --  For X.F(Y,Z), which is shorthand for F(X,Y,Z), First is X
         --  and Rest is Y,Z.

         First : constant Ada_Tree := Actuals.Subtrees (1);

         Rest : constant Ada_Tree :=
           new Ada_Tree_Rec'
             (Kind         => An_Association_List,
             Subtree_Count => Actuals.Subtree_Count - 1,
             Sloc          => Asis.Text.Nil_Span,
             Subtrees      => Actuals.Subtrees (2 .. Actuals.Subtree_Count));

      begin
         if Label_Names.Subtree_Count /= 0 then
            raise Program_Error
              with "labeled prefix calls not yet implemented";
         end if;

         --  ???Work around possible bug in Function_Call_Parameters. Not sure
         --  why Callee would be A_Selected_Component.

         if Callee.Kind /= A_Selected_Component then
            Subtree_To_Ada (First, Cur_Level + 1, Index_In_Parent => 1);
            Buffered_Output.Put (".");
         end if;
         Subtree_To_Ada (Callee, Cur_Level + 1, Index_In_Parent => 2);
         Subtrees_To_Ada (Rest, Pre => " (", Between => ", ", Post => ")");
      end Prefix_Notation_Call;

      procedure Maybe_Blank_Line;

      procedure Maybe_Blank_Line is
         Insert_Blank_Line_Before : Boolean := False;

      begin
         if Tree.Kind = A_Compilation_Unit then
            Insert_Blank_Line_Before := True;
            --  This is to handle:
            --     ---- No blank line after comment.
            --     package P is ...
            --  which otherwise wrongly indents:
            --     ----
            --        package P is ...
            --  ???There should be a better way to fix this.
         end if;

         if
           Tree.Kind in
             An_Ordinary_Type_Declaration |
            --  ???(if rec etc)A_Record_Type_Definition
            --  A_Derived_Record_Extension_Definition

               A_Task_Type_Declaration |
               A_Protected_Type_Declaration |
               A_Single_Task_Declaration |
               A_Single_Protected_Declaration |
               A_Procedure_Body_Declaration |
               A_Function_Body_Declaration |
               A_Package_Declaration | -- ???(non lib unit)
               A_Package_Body_Declaration |
               A_Task_Body_Declaration |
               A_Protected_Body_Declaration |
               An_Entry_Body_Declaration |
               A_Generic_Procedure_Declaration |
               A_Generic_Function_Declaration |
               A_Generic_Package_Declaration |
               An_Enumeration_Type_Definition | --???(if big)
               A_Loop_Statement |
               A_While_Loop_Statement |
               A_For_Loop_Statement |
               A_Block_Statement |
               An_Extended_Return_Statement |
               An_Accept_Statement |
               A_Selective_Accept_Statement |
               A_Timed_Entry_Call_Statement |
               A_Conditional_Entry_Call_Statement |
               An_Asynchronous_Select_Statement |
               An_If_Path | --???look up to If_Statement, then up to list.
               An_Elsif_Path |
               An_Else_Path |
               A_Case_Path |
               A_Record_Representation_Clause
--           An_Exception_Handler |???

         then
            declare
               Parent : constant Ada_Tree := Parent_Tree;

            begin
               if Parent.Kind in Flat_List_Kinds then
                  if Parent.Subtrees (1) /= Tree then
                     Insert_Blank_Line_Before := True;
                  end if;
               end if;
            end;
         end if;

         if Insert_Blank_Line_Before then
            pragma Assert (Line_Breaks (Last (Line_Breaks)).Hard);
            pragma Assert
              (Point (Out_Buf) =
               Last_Position (Out_Buf) + 1); -- ???Do we need Last_Position?
            pragma Assert
              (Position (Out_Buf, Line_Breaks (Last (Line_Breaks)).Mark) =
               Last_Position (Out_Buf));
            pragma Assert (Lookback (Out_Buf) = NL);
            --  There should already be a hard line break here; we're about to
            --  add another one.

            Append_Line_Break
              (Hard     => True,
               Level    => 0,
               Kind     => Tree.Kind,
               Template => Name_Find ("Maybe_Blank_Line"));
         end if;
      end Maybe_Blank_Line;

      use Asis;
      Index : Query_Index := 1;

      --  Procedures for formatting the various kinds of node that are not
      --  fully covered by Template_Table:

      procedure Do_Accept_Statement;
      procedure Do_Association;
      procedure Do_Block_Statement;
      procedure Do_Compilation_Unit;
      procedure Do_Case_Path;
      procedure Do_Case_Statement;
      procedure Do_Constrained_Array_Definition; -- also generic formal
      procedure Do_Def_Name;
      procedure Do_Extended_Return_Statement;
      procedure Do_Extension_Aggregate;
      procedure Do_Function_Call;
      procedure Do_List;
      procedure Do_Literal;
      procedure Do_Null_Statement;
      procedure Do_Ordinary_Type_Declaration;
      procedure Do_Parameter_Specification; -- also Formal_Object_Declaration
      procedure Do_Pragma;
      procedure Do_Procedure_Call_Statement;
      procedure Do_Qualified_Expression;
      procedure Do_Record_Aggregate;
      procedure Do_Single_Task_Declaration;
      procedure Do_Subtype_Indication;
      procedure Do_Task_Type_Declaration;
      procedure Do_Usage_Name;

      procedure Do_Others; -- anything not listed above

      procedure Do_Accept_Statement is
      begin
         --  If there are no statements or exception handlers, use short
         --  form

         if
           Tree.Subtrees (5).Subtree_Count = 0
           and then Tree.Subtrees (6).Subtree_Count = 0
         then
            Interpret_Template (Labels & "accept !? (~~)~?[ (~;@ ~)]~!!");
         --  The last "!!" generates nothing, but satisfies the
         --  requirement that we use all the subtrees.

         else
            Interpret_Template;
         end if;
      end Do_Accept_Statement;

      procedure Do_Association is
         --  Some have a single name before the "=>", and some have a list
         --  separated by "|".
         Positional_Notation : constant Boolean := (case Tree.Kind is
               when
                 A_Pragma_Argument_Association |
                 A_Parameter_Association |
                 A_Generic_Association =>
                  Tree.Subtrees (1).Kind = Not_An_Element,
               when
                 A_Discriminant_Association |
                 A_Record_Component_Association |
                 An_Array_Component_Association =>
                  Tree.Subtrees (1).Subtree_Count = 0, -- list length 0?
               when others =>
                  False); -- Can't happen

      begin
         if Positional_Notation then
            Interpret_Template ("?~~~!");
         --  The "?~~~" generates nothing.

         else
            Interpret_Template ("?~ ^|@ ~ ^=>[ ~!]");
            --  This is needed because the "[]" is not properly nested with the
            --  "?~~~".
         end if;
      end Do_Association;

      procedure Do_Block_Statement is
      begin
         --  If Block_Declarative_Items is empty, leave off the "declare"

         if Tree.Subtrees (3).Subtree_Count = 0 then
            Interpret_Template (Labels & "?~~ : ~!" & Handled_Seq ("2"));

         else
            Interpret_Template
              (Labels &
               "?~~ : ~?declare$" &
               "{~;$~;$$}~" &
               Handled_Seq ("2"));
         end if;
      end Do_Block_Statement;

      use Buffered_Output;

      procedure Do_Compilation_Unit is
      begin
         if False then -- ???
            Put ("--  \1 = \2", "Unit_Kind", Capitalize (Tree.Unit_Kind'Img));
            Interpret_Template ("$", Subtrees => Empty_Tree_Array);
            Put
              ("--  \1 = \2",
               "Unit_Class",
               Capitalize (Tree.Unit_Class'Img));
            Interpret_Template ("$", Subtrees => Empty_Tree_Array);
            Put
              ("--  \1 = \2",
               "Unit_Origin",
               Capitalize (Tree.Unit_Origin'Img));
            Interpret_Template ("$", Subtrees => Empty_Tree_Array);
            Interpret_Template ("$", Subtrees => Empty_Tree_Array);
         end if;
         Subtrees_To_Ada
           (Tree.Subtrees (1),
            Pre     => "",
            Between => ";$",
            Post    => ";$$");
         --  If it's a subunit, we need "separate (Parent.Name)"

         if Tree.Unit_Kind in A_Subunit then
            declare
               N    : constant String := Get_Name_String (Tree.Unit_Full_Name);
               Last : Positive        := N'Last;

            begin
               --  Determine parent name by searching for the last '.'

               while N (Last) /= '.' loop
                  Last := Last - 1;
               end loop;
               Last := Last - 1;

               Put ("separate (\1)", N (1 .. Last));
               Interpret_Template ("$", Subtrees => Empty_Tree_Array);
            end;
         end if;

         case Tree.Unit_Class is
            when A_Private_Declaration =>
               Put ("private ");

            when
              A_Public_Declaration |
              A_Public_Body |
              A_Public_Declaration_And_Body |
              A_Private_Body |
              A_Separate_Body =>
               null;

            when Not_A_Class =>
               raise Program_Error;
         end case;

         Subtree_To_Ada
           (Tree.Subtrees (2),
            Cur_Level + 1,
            Index_In_Parent => 2);
         Put (";");
         Interpret_Template ("$", Subtrees => Empty_Tree_Array);
         Subtrees_To_Ada
           (Tree.Subtrees (3),
            Pre     => "",
            Between => ";$",
            Post    => ";$");
      end Do_Compilation_Unit;

      procedure Do_Case_Path is
         Stms : constant Ada_Tree := Tree.Subtrees (2);

      begin
         --  If the statement list is a single block statement that starts on
         --  the same line as the "when", then we assume the user wants to keep
         --  it that way. For example:
         --
         --     when Upper_Case => Upper_Case_Case : begin

         if
           Stms.Subtree_Count = 1
           and then Stms.Subtrees (1).Kind = A_Block_Statement
           and then Stms.Subtrees (1).Sloc.First_Line = Tree.Sloc.First_Line
         then
            Interpret_Template ("when ?[~ |@ ~]~ => " & "?~~;$~");

         else
            Interpret_Template;
         end if;
      end Do_Case_Path;

      procedure Do_Case_Statement is
         --  If all the "when"s appear in the same column as "case", then we
         --  assume that's what the user intended, and avoid indenting the
         --  "when"s.

         Case_Col : constant Positive := Tree.Sloc.First_Column;
         --  Column in which "case" appears
         Whens_Col : Positive :=
           Tree.Subtrees (3).Subtrees (1).Sloc.First_Column;
      --  Column in which all the "when"s appear, if they're all the same

      begin
         for W of Tree.Subtrees (3).Subtrees loop
            if W.Sloc.First_Column /= Whens_Col then
               Whens_Col := Positive'Last; -- not all the same
            end if;
         end loop;

         if Case_Col = Whens_Col and then Case_Col /= 1 then
            Interpret_Template (Labels & "case[ !]@ is$" & "!" & "end case");

         else
            Interpret_Template;
         end if;
      end Do_Case_Statement;

      procedure Do_Constrained_Array_Definition is
      begin
         --  ???Following is wrong for multi-dim arrays
         --  (only some indices need "range")
         --  And it doesn't work anyway.
         case Tree.Subtrees (1).Subtrees (1).Kind is
            when A_Range_Attribute_Reference | A_Simple_Expression_Range =>
               Interpret_Template ("array (?range ~, range ~~) of !");

            when others =>
               Interpret_Template ("array (?~, ~~) of !");
         end case;
      end Do_Constrained_Array_Definition;

      procedure Do_Def_Name is
      begin
         if Tree.Kind = A_Defining_Expanded_Name then
            Interpret_Template ("![.!]");

         else
            Put ("\1", Get_Name_String (Tree.Def_Name));
         end if;
      end Do_Def_Name;

      procedure Do_Extended_Return_Statement is
      begin
         --  If there are no statements or exception handlers, use short
         --  form

         if
           Tree.Subtrees (2).Subtree_Count = 0
           and then Tree.Subtrees (3).Subtree_Count = 0
         then
            Interpret_Template ("return !!!");
         --  The last "!!" generates nothing, but satisfies the
         --  requirement that we use all the subtrees.

         else
            Interpret_Template;
         end if;
      end Do_Extended_Return_Statement;

      procedure Do_Extension_Aggregate is
      begin
         if Tree.Subtrees (2).Subtree_Count = 0 then
            Interpret_Template ("(! with$" & "{null record)}!");

         else
            Interpret_Template;
         end if;
      end Do_Extension_Aggregate;

      type Precedence_Level is range 1 .. 7;
      function Precedence (Expr : Ada_Tree) return Precedence_Level;

      function Precedence (Expr : Ada_Tree) return Precedence_Level is
      begin
         case Expr.Kind is
--  ???Don't treat membership tests as operators, for now
--            when An_In_Membership_Test | A_Not_In_Membership_Test =>
--               return 1;
            when An_And_Then_Short_Circuit | An_Or_Else_Short_Circuit =>
               return 2;

            when A_Function_Call =>
               --  Binary operator using operator notation

               if
                 Expr.Subtrees (3).Kind /= An_Is_Prefix_Call
                 and then Expr.Subtrees (2).Subtree_Count /= 1
               then
                  pragma Assert
                    (Expr.Subtrees (4).Kind /= An_Is_Prefix_Notation);
                  pragma Assert (Expr.Subtrees (2).Subtree_Count = 2);
                  case Expr.Subtrees (1).Kind is
                     when An_And_Operator | An_Or_Operator | An_Xor_Operator =>
                        return 2; -- same as 'and then' and 'or else'

                     when
                       An_Equal_Operator |
                       A_Not_Equal_Operator |
                       A_Less_Than_Operator |
                       A_Less_Than_Or_Equal_Operator |
                       A_Greater_Than_Operator |
                       A_Greater_Than_Or_Equal_Operator =>
                        return 3;

                     when
                       A_Plus_Operator |
                       A_Minus_Operator |
                       A_Concatenate_Operator =>
                        return 4;

                     when
                       A_Multiply_Operator |
                       A_Divide_Operator |
                       A_Mod_Operator |
                       A_Rem_Operator =>
                        return 5;

                     when An_Exponentiate_Operator =>
                        return 6;

                     when others =>
                        raise Program_Error;
                  end case;

               --  Unary operator or normal function-call notation

               else
                  return 7;
               end if;

            --  Assume anything else is a primary (highest precedence)

            when others =>
               return 7;
         end case;
      end Precedence;

      function Get_Arg (Expr : Ada_Tree; N : Query_Index) return Ada_Tree;

      function Get_Arg (Expr : Ada_Tree; N : Query_Index) return Ada_Tree is
         Assoc : constant Ada_Tree := Expr.Subtrees (2).Subtrees (N);
         pragma Assert (Assoc.Kind = A_Parameter_Association);
         function Is_Positional
           (Assoc : Ada_Tree) return Boolean is
           (Assoc.Subtrees (1).Kind = Not_An_Element);
         pragma Assert (Is_Positional (Assoc));

      begin
         return Assoc.Subtrees (2);
      end Get_Arg;

      function Make_Op (Expr : Ada_Tree) return Ada_Tree;
      --  Create operator node. This is a separate function
      --  to reduce stack usage (for example long strings of
      --  "&" can cause deep recursion).

      function Make_Op (Expr : Ada_Tree) return Ada_Tree is
      begin
         return Result : constant Ada_Tree := Make (An_Identifier) do
            case Expr.Kind is
               when A_Function_Call =>
                  declare
                     Q_Op_Sym : constant String :=
                       Get_Name_String (Expr.Subtrees (1).Ref_Name);

                  begin
                     Result.Ref :=
                       Name_Find (Q_Op_Sym (2 .. Q_Op_Sym'Last - 1));
                     --  Strip off quotes
                  end;

               when An_And_Then_Short_Circuit =>
                  Result.Ref := Name_And_Then;

               when An_Or_Else_Short_Circuit =>
                  Result.Ref := Name_Or_Else;

               when others =>
                  raise Program_Error;
            end case;
            Result.Ref_Name := Result.Ref;
         end return;
      end Make_Op;

      procedure Do_Unary_Operator (Expr : Ada_Tree);

      procedure Do_Binary_Operator
        (Expr      : Ada_Tree;
         Is_Right  : Boolean;
         Cur_Level : Nesting_Level);
      --  Also handles some things that look like operators, like "and then".
      --  Is_Right is True if Expr is the right-hand argument of an outer
      --  binary operator. Otherwise (Expr is the left-hand argument, or
      --  Expr's parent is something else, like a parenthesized expression),
      --  Is_Right is False.

      function Is_Bin_Op (Expr : Ada_Tree) return Boolean;

      procedure Do_Unary_Operator (Expr : Ada_Tree) is
         Op       : constant Ada_Tree       := Make_Op (Expr);
         Arg1     : constant Ada_Tree       := Get_Arg (Expr, 1);
         Subtrees : constant Ada_Tree_Array := (Op, Arg1);

      begin
         if
           Expr.Subtrees (1).Kind in
             A_Unary_Plus_Operator |
               A_Unary_Minus_Operator
         then
            Interpret_Template ("!!", Subtrees (1 .. 2));

         else
            Interpret_Template ("! !", Subtrees (1 .. 2));
         end if;
      end Do_Unary_Operator;

      function Is_Bin_Op (Expr : Ada_Tree) return Boolean is
      begin
         case Expr.Kind is
            when A_Function_Call =>
               return
                 Expr.Subtrees (3).Kind /= An_Is_Prefix_Call
                 and then Expr.Subtrees (2).Subtree_Count = 2;

            when An_And_Then_Short_Circuit | An_Or_Else_Short_Circuit =>
               return True;

            when others =>
               return False;
         end case;
      end Is_Bin_Op;

      procedure Do_Binary_Operator
        (Expr      : Ada_Tree;
         Is_Right  : Boolean;
         Cur_Level : Nesting_Level)
      is
         Is_Short_C : constant Boolean :=
           Expr.Kind in An_And_Then_Short_Circuit | An_Or_Else_Short_Circuit;
         Is_Expon : constant Boolean := -- True for "**"
           (Expr.Kind in A_Function_Call
            and then Expr.Subtrees (1).Kind = An_Exponentiate_Operator);
         Op : constant Ada_Tree := Make_Op (Expr);
         Arg1, Arg2  : Ada_Tree;
         Arg1_Higher : Boolean; -- Arg1 is higher precedence than Expr

      --  Calculate template fragments for the args (Arg1/2_T), that indent
      --  if the arg is a higher precedence binary operator than the whole
      --  expression.

      --  Start of processing for Do_Binary_Operator

      begin
         if Is_Short_C then
            Arg1 := Expr.Subtrees (1);
            Arg2 := Expr.Subtrees (2);

         else -- function call
            Arg1 := Get_Arg (Expr, 1);
            Arg2 := Get_Arg (Expr, 2);
         end if;

         --  The arguments can't have lower precedence than the expression as a
         --  whole; that's what precedence means -- you need parens to put a
         --  "+" inside a "*". The right-hand argument can't have equal
         --  precedence, because Ada has no right-associative binary operators.

         pragma Assert (Precedence (Arg1) >= Precedence (Expr));
         pragma Assert (Precedence (Arg2) > Precedence (Expr));

         Arg1_Higher := Precedence (Arg1) > Precedence (Expr);

         --  The recursive calls to Do_Binary_Operator below bypass the normal
         --  recursion via Subtree_To_Ada, so we need to pass along the
         --  Cur_Level to Interpret_Template. When we reach something that's
         --  not a binary op, we switch back to the normal recursion via
         --  Interpret_Template on the Arg. We split lines after the operator
         --  symbol, as in:
         --     Some_Long_Thing +
         --     Some_Other_Long_Thing
         --  except in the case of short circuits:
         --     Some_Long_Thing
         --     and then Some_Other_Long_Thing
         --  All operators are surrounded by blanks, except for "**":
         --     Max : constant := 2**31 - 1;

         if Is_Bin_Op (Arg1) then
            if Is_Right and then Arg1_Higher then
               Interpret_Template ("[", Empty_Tree_Array, Cur_Level);
            end if;
            Do_Binary_Operator
              (Arg1,
               Is_Right  => Is_Right,
               Cur_Level => Cur_Level + (if Arg1_Higher then
                    1
                 else
                    0));
            if Is_Right and then Arg1_Higher then
               Interpret_Template ("]", Empty_Tree_Array, Cur_Level);
            end if;

         else
            Interpret_Template
              ("!",
               Subtrees => (1 => Arg1),
               Cur_Level => Cur_Level);
         end if;

         if Is_Short_C then
            Interpret_Template ("@", Empty_Tree_Array, Cur_Level);
         end if;
         Interpret_Template ((if Is_Expon then
               "!"
            else
               " ! "), -- no blanks for "**"
               Subtrees => (1 => Op), Cur_Level => Cur_Level);
         if not Is_Short_C then
            Interpret_Template ("@", Empty_Tree_Array, Cur_Level);
         end if;

         if Is_Bin_Op (Arg2) then
            Interpret_Template ("[", Empty_Tree_Array, Cur_Level + 1);
            Do_Binary_Operator
              (Arg2,
               Is_Right  => True,
               Cur_Level => Cur_Level + 1);
            Interpret_Template ("]", Empty_Tree_Array, Cur_Level + 1);

         else
            Interpret_Template
              ("!",
               Subtrees => (1 => Arg2),
               Cur_Level => Cur_Level + 1);
         end if;
      end Do_Binary_Operator;

      procedure Do_Function_Call is
      begin
         --  Note: Is_Prefix_Notation is for Object.Operation(...)
         --  notation, whereas Is_Prefix_Call is for anything that's not an
         --  operator notation call. Thus Is_Prefix_Call is True for "&"(X,
         --  Y), and False for X&Y.

         if Tree.Subtrees (4).Kind = An_Is_Prefix_Notation then
            pragma Assert (Tree.Subtrees (3).Kind = An_Is_Prefix_Call);
            Prefix_Notation_Call
              (Label_Names => Empty (A_Defining_Name_List),
               Callee      => Tree.Subtrees (1),
               Actuals     => Tree.Subtrees (2));

         --  Determine whether to use operator notation, like X+Y instead of
         --  "+"(X,Y). We can use operator notation if it's an operator
         --  call, and the argument(s) are in positional notation (not
         --  named). ???We must use operator notation for "/=", to work
         --  around compiler bug. In some cases, "/="(X, Y) doesn't work (on
         --  access types?), so we generate (X /= Y) instead.

         --  We don't want to translate
         --  "&" (STRING'("AB"), STRING'("CDEF"))(5) /= CHARACTER'('E')
         --  into
         --  ((STRING'("AB") & STRING'("CDEF"))(5) /= CHARACTER'('E'))
         --  because an operator-notation call is not a name, and
         --  therefore cannot be used as the prefix of an indexed
         --  component.

         elsif Tree.Subtrees (3).Kind = An_Is_Prefix_Call then
            Interpret_Template; -- normal "F (X)" notation

         --  Operator notation:

         else
            pragma Assert
              (Tree.Subtrees (1).Kind in Flat_Operator_Symbol_Kinds);
            pragma Assert (Tree.Subtrees (2).Subtree_Count in 1 .. 2);

            --  Unary operator

            if Tree.Subtrees (2).Subtree_Count = 1 then
               Do_Unary_Operator (Tree);

            --  Binary operator

            else
               Do_Binary_Operator
                 (Tree,
                  Is_Right  => False,
                  Cur_Level => Cur_Level);
            end if;
         end if;
      end Do_Function_Call;

      procedure Do_List is
      begin
         Subtrees_To_Ada (Tree, Pre => "", Between => "$", Post => "");
      end Do_List;

      procedure Do_Literal is
      begin
         pragma Assert (Check_Whitespace);
         Check_Whitespace := False;
         Put ("\1", Get_Name_String (Tree.Lit_Val));
         Check_Whitespace := True;
      end Do_Literal;

      procedure Do_Null_Statement is
      begin
         --  If a label comes at the end of a statement list, as allowed in
         --  Ada 2012, ASIS inserts an extra implicit null statement to hang
         --  the label off of. We don't want to print that statement,
         --  because it wasn't in the source code. We can detect such
         --  implicit null statements by checking for a nil Sloc. We also
         --  need to suppress the ";" that comes after the implicit 'null',
         --  which is the purpose of Implicit_Null_Statement_Kludge. We set
         --  that flag True here, and the very next template character seen
         --  by Interpret_Template will be that ";", so Interpret_Template
         --  will suppress the ";" and reset Implicit_Null_Statement_Kludge
         --  to False.

         if
           Tree.Subtrees (1).Subtree_Count /= 0
           and then Asis.Text.Is_Nil (Tree.Sloc)
         then
            Interpret_Template (Labels);
            Implicit_Null_Statement_Kludge := True;

         else
            Interpret_Template;
         end if;
      end Do_Null_Statement;

      procedure Do_Ordinary_Type_Declaration is
      begin
         if
           Tree.Subtrees (3).Kind in
             A_Derived_Record_Extension_Definition |
               A_Record_Type_Definition |
               A_Tagged_Record_Type_Definition
         then
            Interpret_Template ("type !! is !" & Aspects);
         --  Record_Definition will take care of new lines

         else
            Interpret_Template;
         end if;
      end Do_Ordinary_Type_Declaration;

      procedure Do_Others is
      begin
         if Template_Table (Tree.Kind) = null then
            Put ("null templ:\1", Image (Tree.Kind));
            Subtrees_To_Ada (Tree, Pre => "{", Between => "|", Post => "}");

         else
            Interpret_Template;
         end if;
      end Do_Others;

      procedure Do_Parameter_Specification is
      begin
         Subtrees_To_Ada
           (Tree.Subtrees (Index),
            Pre     => "",
            Between => ",@ ",
            Post    => "");
         Interpret_Template (" ^: ", Subtrees => Empty_Tree_Array);

         case Tree.Kind is
            when A_Parameter_Specification =>
               Index := Index + 1;

               if
                 Tree.Subtrees (Index).Kind /= Not_An_Element
               then -- "aliased"
                  Subtree_To_Ada (Tree.Subtrees (Index), Cur_Level + 1, Index);
                  Put (" ");
               end if;

            when A_Formal_Object_Declaration =>
               null; -- A_Formal_Object_Declaration doesn't have "aliased"

            when others =>
               raise Program_Error;
         end case;

         case Tree.Mode is
            when Not_A_Mode =>
               raise Program_Error;

            when A_Default_In_Mode =>
               null;

            when An_In_Mode =>
               if False then
                  Put ("in ");
               end if; -- ???
            --  Should be an option for explicit-in, no-in, leave-it-alone,
            --  &c.

            when An_Out_Mode =>
               Put ("out ");

            when An_In_Out_Mode =>
               Put ("in out ");
         end case;

         Index := Index + 1;

         if Tree.Subtrees (Index).Kind /= Not_An_Element then -- "not null"
            Subtree_To_Ada (Tree.Subtrees (Index), Cur_Level + 1, Index);
            Put (" ");
         end if;

         Index := Index + 1;
         Subtree_To_Ada (Tree.Subtrees (Index), Cur_Level + 1, Index);

         Index := Index + 1;
         if Tree.Subtrees (Index).Kind /= Not_An_Element then
            Interpret_Template
              (" ^:=[ !]",
               Subtrees => (1 => Tree.Subtrees (Index)));
         end if;
      end Do_Parameter_Specification;

      procedure Do_Pragma is
      begin
         Put ("pragma \1", Get_Name_String (Tree.Pragma_Name));
         Interpret_Template ("?[ (~,@ ~)]~");
      end Do_Pragma;

      procedure Do_Procedure_Call_Statement is
      begin
         if Tree.Subtrees (4).Kind = An_Is_Prefix_Notation then
            Prefix_Notation_Call
              (Label_Names => Tree.Subtrees (1),
               Callee      => Tree.Subtrees (2),
               Actuals     => Tree.Subtrees (3));

         else
            Interpret_Template;
         end if;
      end Do_Procedure_Call_Statement;

      procedure Do_Qualified_Expression is
      begin
         if
           Tree.Subtrees (2).Kind in
             A_Record_Aggregate |
               An_Extension_Aggregate |
               A_Positional_Array_Aggregate |
               A_Named_Array_Aggregate
         then
            Interpret_Template ("!'[!]");
         --  If the thing after the ' is an aggregate, we leave out
         --  the parentheses here, because the aggregate will insert
         --  them. We want T'(X, Y, Z), not T'((X, Y, Z)).

         else
            Interpret_Template;
         end if;
      end Do_Qualified_Expression;

      procedure Do_Record_Aggregate is
      begin
         if Tree.Subtrees (1).Subtree_Count = 0 then
            Interpret_Template ("(null record)!");

         else
            Interpret_Template;
         end if;
      end Do_Record_Aggregate;

      procedure Do_Single_Task_Declaration is
      begin
         if
           Tree.Subtrees (3).Subtree_Count = 0
           and then Tree.Subtrees (4).Subtree_Count = 0
         then
            Interpret_Template ("task !" & Aspects & "!!");

         else
            Interpret_Template;
         end if;
      end Do_Single_Task_Declaration;

      procedure Do_Subtype_Indication is
      begin
         case Tree.Subtrees (4).Kind is
            when A_Range_Attribute_Reference | A_Simple_Expression_Range =>
               Interpret_Template ("?~~ ~?~~ ~!? range ~~~");

            when others =>
               Interpret_Template ("?~~ ~?~~ ~!? ~~~");
         end case;
      end Do_Subtype_Indication;

      procedure Do_Task_Type_Declaration is
      begin
         --  For task [type] declarations, use short form if possible

         if
           Tree.Subtrees (4).Subtree_Count = 0
           and then Tree.Subtrees (5).Subtree_Count = 0
         then
            Interpret_Template ("task type !!" & Aspects & "!!");

         else
            Interpret_Template;
         end if;
      end Do_Task_Type_Declaration;

      procedure Do_Usage_Name is
      begin
         Put ("\1", Get_Name_String (Tree.Ref_Name));
      end Do_Usage_Name;

   --  Start of processing for Subtree_To_Ada

   begin
      Append (Tree_Stack, Tree); -- push

      Maybe_Blank_Line;

      case Tree.Kind is
         when A_Compilation_Unit =>
            Do_Compilation_Unit;

         when Def_Names =>
            Do_Def_Name;

         when Usage_Names =>
            Do_Usage_Name;

         when An_Integer_Literal | A_Real_Literal | A_String_Literal =>
            Do_Literal;

         when Flat_Pragma_Kinds =>
            Do_Pragma;

         when A_Null_Statement =>
            Do_Null_Statement;

         when An_Ordinary_Type_Declaration =>
            Do_Ordinary_Type_Declaration;

         when A_Procedure_Call_Statement =>
            Do_Procedure_Call_Statement;

         when A_Function_Call =>
            Do_Function_Call;

         when An_And_Then_Short_Circuit | An_Or_Else_Short_Circuit =>
            Do_Binary_Operator
              (Tree,
               Is_Right  => False,
               Cur_Level => Cur_Level);

         when A_Task_Type_Declaration =>
            Do_Task_Type_Declaration;

         when A_Single_Task_Declaration =>
            Do_Single_Task_Declaration;

         when
           A_Pragma_Argument_Association |
           A_Discriminant_Association |
           A_Record_Component_Association |
           An_Array_Component_Association |
           A_Parameter_Association |
           A_Generic_Association =>
            Do_Association;

         when A_Block_Statement =>
            Do_Block_Statement;

         when A_Subtype_Indication =>
            Do_Subtype_Indication;

         when A_Case_Path =>
            Do_Case_Path;

         when A_Case_Statement =>
            Do_Case_Statement;

         when
           A_Constrained_Array_Definition |
           A_Formal_Constrained_Array_Definition =>
            Do_Constrained_Array_Definition;

         when An_Extended_Return_Statement =>
            Do_Extended_Return_Statement;

         when An_Accept_Statement =>
            Do_Accept_Statement;

         when A_Qualified_Expression =>
            Do_Qualified_Expression;

         when A_Record_Aggregate =>
            Do_Record_Aggregate;

         when An_Extension_Aggregate =>
            Do_Extension_Aggregate;

         when A_Parameter_Specification | A_Formal_Object_Declaration =>
            Do_Parameter_Specification;

         when Flat_List_Kinds =>
            Do_List;

         when others =>
            Do_Others;
      end case;

      Delete_Last (Tree_Stack); -- pop
   end Subtree_To_Ada;

   procedure Insert_Comments_And_Blank_Lines
     (Src_Tokens, Out_Tokens : Scanner.Token_Vectors.Vector;
      Do_Inserts             : Boolean);
   --  Src_Tokens is the tokens from the original source file.
   --  Out_Tokens is the newly-generated tokens.
   --  Out_Buf contains the corresponding characters to Out_Tokens.
   --  This procedure is called twice, first with Do_Inserts True, then
   --  False:
   --  The first time, Out_[Tokens|Buf] doesn't contain any comments; they are
   --  inserted into the output from Src_Tokens.
   --  The second time, they do, and we check that they match the comments in
   --  Src_Tokens. This second pass is not strictly necessary; it's purpose is
   --  to make sure that we crash rather than garbling the user's program.

   procedure Insert_Comments_And_Blank_Lines
     (Src_Tokens, Out_Tokens : Scanner.Token_Vectors.Vector;
      Do_Inserts             : Boolean)
   is
      use Scanner;

      function Comment_Text (Name : Name_Id) return String;
      --  Name is the normalized text of a comment. This normally returns the
      --  text unmodified, except that in some cases we change "-- XXX" to
      --  "--  XXX" (inserting a second blank).

      function Match (Tok1, Tok2 : Token) return Boolean;
      --  True if the tokens have the same kind and same text, except that the
      --  matching is case insensitive for identifiers, reserved words, and
      --  string literals that could be operator symbols. Also, allow an
      --  End_Of_Line_Comment to match a Whole_Line_Comment if Do_Inserts
      --  is False. The source locations are ignored.

      function Next_Lexeme
        (Tokens : Token_Vectors.Vector;
         Index  : Token_Index)
        return Token;
      --  Returns the next token that is not a blank line or comment

      function Prev_Lexeme
        (Tokens : Token_Vectors.Vector;
         Index  : Token_Index)
        return Token;
      --  Returns the previous token that is not a blank line or comment

      procedure Move_Past_Char;
      procedure Move_Past_Out_Tok;

      procedure Insert_End_Of_Line_Comment;
      --  Found an End_Of_Line_Comment comment; copy it to the buffer. If it is
      --  too long to fit on the line, turn it into a Whole_Line_Comment,
      --  taking care to indent.

      --  Note that the Subtree_To_Ada pass already inserted indentation, so we
      --  mostly keep the indentation level at zero. The exception is comments,
      --  which Subtree_To_Ada didn't see. For comments, we set the indentation
      --  to that of the surrounding code, and Move_Past_Out_Tok will set it
      --  back to zero when it prints a nonblank character.

      procedure Insert_Whole_Line_Comment;
      --  Found a Whole_Line_Comment; copy it to the buffer, taking care to
      --  indent, except that if the comment starts in column 1, we assume the
      --  user wants to keep it that way.

      procedure Insert_Declare_Or_Private (Declare_Or_Private : String)
        with Pre => Declare_Or_Private in "declare" | "private";
      --  If a block statement has no declarations, the earlier passes don't
      --  insert "declare", whether or not it was in the source code. If
      --  Do_Inserts is True, and there is a comment, this re-inserts
      --  "declare" before the comment, to avoid messing up the formatting.
      --  Similarly for "private [possible comment] end".

      function Comment_Text (Name : Name_Id) return String is
         S : constant String := Get_Name_String (Name);

      begin
         pragma Assert (S'First = 1);
         if Has_Suffix (S, Suffix => "--") then
            null; -- Don't modify copyright headers, etc.

         elsif Has_Prefix (S, Prefix => "--  ") then
            null; -- Don't modify if we already have extra blank

         elsif not Has_Prefix (S, Prefix => "-- ") then
            null; -- Don't modify SPARK comments, etc.

         else
            return "--  " & S (4 .. S'Last);
         end if;

         return S;
      end Comment_Text;

      function Match (Tok1, Tok2 : Token) return Boolean is
      begin
         if Tok1.Kind = Tok2.Kind then
            case Tok1.Kind is
               when Nil | End_Of_Line =>
                  raise Program_Error;

               when
                 Start_Of_Input |
                 End_Of_Input |
                 Lexeme |
                 Whole_Line_Comment |
                 End_Of_Line_Comment |
                 Blank_Line =>
                  if Tok1.Normalized = Tok2.Normalized then
                     return True;

                  elsif
                    not Do_Inserts and then Tok1.Kind = Whole_Line_Comment
                  then
                     --  Allow for the fact that we might have inserted an
                     --  extra blank. ???This won't work when we're formatting
                     --  comments more agressively.
                     declare
                        T1 : constant String :=
                          Get_Name_String (Tok1.Normalized);
                        T2 : constant String :=
                          Get_Name_String (Tok2.Normalized);

                     begin
                        return
                          T1 (1 .. 3) = "-- "
                          and then T2 (1 .. 4) = "--  "
                          and then T1 (4 .. T1'Last) = T2 (5 .. T2'Last);
                     end;

                  else
                     return False;
                  end if;

               when Identifier | Reserved_Word =>
                  return Tok1.Normalized = Tok2.Normalized;

               when String_Literal =>
                  if Is_Op_Sym_With_Letters (Tok1.Normalized) then
                     return Tok1.Normalized = Tok2.Normalized;

                  else
                     return Tok1.Text = Tok2.Text;
                  end if;
            end case;

         elsif
           not Do_Inserts
           and then Tok1.Kind = End_Of_Line_Comment
           and then Tok2.Kind = Whole_Line_Comment
         then
            if Tok1.Normalized = Tok2.Normalized then
               return True;

            elsif
              Comment_Text (Tok1.Normalized) =
              Get_Name_String (Tok2.Normalized)
            then
               return True;

            else
               return False;
            end if;

         else
            return False;
         end if;
      end Match;

      function Next_Lexeme
        (Tokens : Token_Vectors.Vector;
         Index  : Token_Index)
        return Token
      is
         X : Token_Index := Index + 1;

      begin
         while Tokens (X).Kind in Blank_Line | Comment_Kind loop
            X := X + 1;
         end loop;

         return Tokens (X);
      end Next_Lexeme;

      function Prev_Lexeme
        (Tokens : Token_Vectors.Vector;
         Index  : Token_Index)
        return Token
      is
         X : Token_Index := Index - 1;

      begin
         while Tokens (X).Kind in Blank_Line | Comment_Kind loop
            X := X - 1;
         end loop;

         return Tokens (X);
      end Prev_Lexeme;

      Src_Index, Out_Index : Token_Index := 2;
      --  Skip the first Start_Of_Input token, which is just a sentinel

      Src_Tok, Out_Tok : Token;

      Line_Breaks : Line_Break_Vector renames Syntax_Line_Breaks;
      --  Line breaks used for indenting whole-line comments

      --  ???
      EOL_Line_Breaks : Line_Break_Vector renames Enabled_Line_Breaks;
--  EOL_Line_Breaks : Line_Break_Vector renames Nonblank_Line_Breaks;
      --  Line breaks used for indenting end-of-line comments

      Cur_Line : Line_Break_Index := 2;
      Cur_Line_Break : Line_Break; -- ???Not used!
      EOL_Cur_Line : Line_Break_Index := 2; -- for end-of-line comments
      EOL_Cur_Line_Break : Line_Break; -- ???Not used!
      Cur_Tab_Index : Tab_Index := 1;
      Cur_Tab       : Tab_Rec   := Tabs (Cur_Tab_Index);

      procedure Move_Past_Char is
      begin
         pragma Assert
           (Point (Out_Buf) <= Position (Out_Buf, Cur_Line_Break.Mark));

         --  Step past Line_Breaks at the current position

         while Cur_Line <= Last_Index (Line_Breaks)
         and then At_Point (Out_Buf, Cur_Line_Break.Mark)
         loop
            Cur_Line := Cur_Line + 1;
            if Cur_Line <= Last_Index (Line_Breaks) then -- ???Kludgy
               Cur_Line_Break := Line_Breaks (Cur_Line);
            end if;
         end loop;

         --  Step past EOL_Line_Breaks at the current position

         while EOL_Cur_Line <= Last_Index (EOL_Line_Breaks)
         and then At_Point (Out_Buf, EOL_Cur_Line_Break.Mark)
         loop
            EOL_Cur_Line := EOL_Cur_Line + 1;
            if EOL_Cur_Line <= Last_Index (EOL_Line_Breaks) then
               EOL_Cur_Line_Break := EOL_Line_Breaks (EOL_Cur_Line);
            end if;
         end loop;

         --  Step past Tabs at the current position

         pragma Assert (Point (Out_Buf) <= Position (Out_Buf, Cur_Tab.Mark));

         while At_Point (Out_Buf, Cur_Tab.Mark) loop
            Cur_Tab_Index := Cur_Tab_Index + 1;
            Cur_Tab       := Tabs (Cur_Tab_Index);
         end loop;

         --  Step past character

         Move_Forward (Out_Buf);
      end Move_Past_Char;

      procedure Move_Past_Out_Tok is
      begin
         loop
            Move_Past_Char;
            exit when At_Point (Out_Buf, Out_Tok.Sloc.Lastx);
         end loop;
      end Move_Past_Out_Tok;

      procedure Insert_End_Of_Line_Comment is
         Indentation : Natural := 0;

      begin
         pragma Assert (EOL_Cur_Line > 1);
         Indentation := EOL_Line_Breaks (EOL_Cur_Line - 1).Indentation;

         --  Insert a blank, except that if we're just before a blank, move
         --  past it, in case it's followed by NL, so we won't add a new NL
         --  below.

         if
           not At_Point (Out_Buf, EOL_Line_Breaks (EOL_Cur_Line).Mark)
           and then Char_At (Out_Buf, Point (Out_Buf)) = ' '
         then
            Move_Past_Char;

         else
            Insert (Out_Buf, ' ');
         end if;
         Insert (Out_Buf, Get_Name_String (Src_Tok.Normalized));

         --  In the usual case, the end-of-line comment is at a natural
         --  line break, like this:
         --      X := X + 1; -- Increment X
         --  so we don't need another one. But if the original was:
         --      X := -- Increment X
         --        X + 1;
         --  we need to add a line break after the comment.

         if not At_Point (Out_Buf, EOL_Line_Breaks (EOL_Cur_Line).Mark) then
            pragma Assert (Cur (Out_Buf) /= NL);
            Cur_Indentation := Indentation;
            Append_Temp_Line_Break;
            Cur_Indentation := 0;
         end if;
         Src_Index := Src_Index + 1;
      end Insert_End_Of_Line_Comment;

      procedure Insert_Whole_Line_Comment is
         function Look_Before return Boolean;
         --  True if we should look before the current location to
         --  determine indentation level for the comment. If the next
         --  lexeme is "begin", for example, we want to indent to the
         --  level of "begin", even though there is probably previous
         --  code more deeply indented.

         function Look_Before return Boolean is
         begin
            if Out_Tok.Kind = End_Of_Input then
               return True;
            end if;

            declare
               X : constant Token := Next_Lexeme (Out_Tokens, Out_Index);

            begin
               return
                 not (X.Normalized = Snames.Name_Begin
                 or else X.Normalized = Snames.Name_When
                 or else X.Normalized = Snames.Name_Elsif
                 or else X.Normalized = Snames.Name_Else);
            end;
         end Look_Before;

         Indentation : Natural := 0;

      --  Start of processing for Insert_Whole_Line_Comment

      begin
         Indentation := Line_Breaks (Cur_Line).Indentation;

         if Cur_Line > 1 and then Look_Before then
            Indentation :=
              Natural'Max
                (Indentation,
                 Line_Breaks (Cur_Line - 1).Indentation);
         end if;

         --  Make sure Indentation is a multiple of PP_Indentation; otherwise
         --  style checking complains "(style) bad column".

         Indentation :=
           (Indentation / Options.PP_Indentation) * Options.PP_Indentation;
         pragma Assert ((Indentation mod Options.PP_Indentation) = 0);

         if Src_Tok.Sloc.Col = 1 then
            Cur_Indentation := 0;

         else
            Cur_Indentation := Indentation;
         end if;
         if
           Src_Tokens (Src_Index - 1).Kind = Blank_Line
           or else (Cur_Line > 1 and then -- Needed???
           Lookback (Out_Buf) /= NL)
         then
            Append_Temp_Line_Break;
         end if;

         loop
            --  ???Handle blank lines here, too?
            Insert (Out_Buf, Comment_Text (Src_Tok.Normalized));
            Src_Index := Src_Index + 1;
            Src_Tok   := Src_Tokens (Src_Index);
            if Src_Tok.Sloc.Col = 1 then
               Cur_Indentation := 0; -- keep comment in column 1

            else
               Cur_Indentation := Indentation;
            end if;
            exit when Src_Tok.Kind /= Whole_Line_Comment;
            Append_Temp_Line_Break;
         end loop;

         --  If the comment occurs in a place in the syntax that doesn't
         --  normally get a line break, we need to add one.

         if Cur (Out_Buf) /= NL then
            Cur_Indentation := Indentation;
            Append_Temp_Line_Break;
         end if;

         Cur_Indentation := 0;
      end Insert_Whole_Line_Comment;

      procedure Insert_Declare_Or_Private (Declare_Or_Private : String) is
         Out_Tok_Pos : constant Positive :=
           Position (Out_Buf, Out_Tok.Sloc.Firstx);
         LB_Pos : constant Positive :=
           Position (Out_Buf, Line_Breaks (Cur_Line).Mark);
         Prev_LB_Pos : constant Positive :=
           Position (Out_Buf, Line_Breaks (Cur_Line - 1).Mark);

      begin
         --  Maybe we should insert "declare" or "private" unconditionally???

         if
           Do_Inserts
           and then Src_Tokens (Src_Index + 1).Kind in
               Comment_Kind | Blank_Line
         then
            --  Either the current or previous line break is just before
            --  "begin" or "end"; that's the indentation we want for "declare"
            --  or "private", respectively. There is one exception: a named
            --  block of the form "Name : begin", we want to insert the declare
            --  before "begin", and we don't care about indentation. ???Better
            --  would be to use indentation of "Name".

            if LB_Pos = Out_Tok_Pos - 1 then
               Cur_Indentation := Line_Breaks (Cur_Line).Indentation;

            elsif Prev_LB_Pos = Out_Tok_Pos - 1 then
               Cur_Indentation := Line_Breaks (Cur_Line - 1).Indentation;

            --  The "one exception" mentioned above

            else
               pragma Assert
                 (Declare_Or_Private = "declare"
                    and then Out_Tokens (Out_Index - 1).Text = Name_Colon
                    and then Out_Tokens (Out_Index - 2).Kind = Identifier);
            end if;

            Append_Temp_Line_Break;
            Insert (Out_Buf, Declare_Or_Private);
            Cur_Indentation := 0;
         end if;
         Src_Index := Src_Index + 1;
      end Insert_Declare_Or_Private;

   --  Start of processing for Insert_Comments_And_Blank_Lines

   begin
      pragma Assert (Cur (Out_Buf) = NL);
      Move_Forward (Out_Buf); -- skip sentinel
      Collect_Enabled_Line_Breaks;
      Cur_Line_Break     := Line_Breaks (Cur_Line);
      EOL_Cur_Line_Break := EOL_Line_Breaks (EOL_Cur_Line);
      Clear (Temp_Line_Breaks);

      --  The two sequences Src_Tokens and Out_Tokens should be identical, with
      --  some exceptions where mismatches are possible. The code below to
      --  insert comments depends on this fact. We step through the two
      --  sequences, copying text into Buffer, and detect any token
      --  mismatch. The allowed mismatches are:
      --
      --     The Out sequence has no comments the first time this is called,
      --     so when we detect a mismatch and the source one is a comment,
      --     that's where we insert the comment. The second time this is
      --     called, the comments must match.
      --
      --     The sequences may have blank lines in different places.
      --
      --     We normalize "end;" to "end Some_Name;"
      --
      --     We normalize single_task_declaration to "task Task_Name;" without
      --     the "is ... end Task_Name", when possible. Similarly for
      --     task_type_declaration.
      --
      --     We normalize by removing "declare" from a block statement with no
      --     declarative items. If there is a comment between "declare" and
      --     "begin", we put the "declare" back in.
      --
      --     We normalize by removing "private" from a package (etc) when there
      --     is nothing in the private part.
      --
      --     We normalize a qualified expression with unnecessary parentheses
      --     containing an aggregate. That is "T'((X, Y, Z))" is normalized to
      --     "T'(X, Y, Z)", where "(X, Y, Z)" is an aggregate. We pretty-much
      --     have to do that, because ASIS provides no way to distinguish these
      --     two forms.
      --
      --     We normalize "X : in T" to "X : T".
      --
      --     Allowed Replacements of Characters (see RM-J.2). We normalize "!"
      --     to "|" when used as a delimiter. The other allowed replacements
      --     (: for # and % for ") are not normalized.
      --
      --  Any other mismatch is considered to be a bug in xml2gnat.

      loop
         Src_Tok := Src_Tokens (Src_Index);
         Out_Tok := Out_Tokens (Out_Index);

         pragma Assert (if Do_Inserts then
               Out_Tok.Kind not in Comment_Kind);

               --  Move into comment area???
         pragma Assert
           (Prev_Lexeme (Out_Tokens, Out_Index).Kind not in
              Blank_Line |
                Comment_Kind);

         --  The order of the if/elsif's below is important in some
         --  cases. Blank lines must be handled late, even if they
         --  match. End_Of_Line_Comments must be handled before blank lines,
         --  because they need to appear at the end of the preceding
         --  line. Whole_Line_Comments must be handled after blank lines,
         --  because the blank line should precede the comment.

         if
           Src_Tok.Kind /= Blank_Line
           and then (Match (Src_Tok, Out_Tok)
           or else (Src_Tok.Normalized = Name_Bang
           and then Out_Tok.Normalized = Name_Bar))
         then
            Move_Past_Out_Tok;

            exit when Src_Tok.Kind = End_Of_Input;
            --  i.e. exit when both Src and Out are at end of input

            Src_Index := Src_Index + 1;
            Out_Index := Out_Index + 1;

         else
            --  Check for "end;" --> "end Some_Name;" case
--???Check next Out token is ";"
            if
              Src_Tok.Text = Name_Semicolon
              and then
                Prev_Lexeme (Src_Tokens, Src_Index).Normalized =
                Snames.Name_End
              and then Out_Tok.Kind in Identifier | String_Literal
            then
               loop -- could be "end A.B.C;"
                  Move_Past_Out_Tok;
                  Out_Index := Out_Index + 1;
                  Out_Tok := Out_Tokens (Out_Index);
                  --  ???Shouldn't have to set Out_Tok here. Either write a
                  --  procedure that sets it every time Out_Index changes, or
                  --  make Out_Tok a function.

                  exit when Out_Tok.Normalized /= Name_Dot;

                  Move_Past_Out_Tok;
                  Out_Index := Out_Index + 1;
                  Out_Tok := Out_Tokens (Out_Index);
                  if Out_Tok.Kind not in Identifier | String_Literal then
                     raise Program_Error;
                  end if;
               end loop;

            --  Check for "declare begin" --> "begin" case, with a possible
            --  comment between "declare" and "begin".

            elsif
              Src_Tok.Normalized = Snames.Name_Declare
              and then Out_Tok.Normalized = Snames.Name_Begin
            then
               pragma Assert
                 (Next_Lexeme (Src_Tokens, Src_Index).Normalized =
                  Snames.Name_Begin);
               Insert_Declare_Or_Private ("declare");

            --  Check for "private end" --> "end" case.

            elsif
              Src_Tok.Normalized = Snames.Name_Private
              and then Out_Tok.Normalized = Snames.Name_End
            then
               pragma Assert
                 (Next_Lexeme (Src_Tokens, Src_Index).Normalized =
                  Snames.Name_End);
               Insert_Declare_Or_Private ("private");

            --  Check for "T'((X, Y, Z))" --> "T'(X, Y, Z)" case

            elsif
              Src_Tok.Text = Name_L_Paren
              and then Prev_Lexeme (Src_Tokens, Src_Index).Text = Name_L_Paren
               --???Also check that the one before that is a tick!
               --  We could count nesting levels of qualified expressions,
               --  and assert that the removed parens match.

            then
               Src_Index := Src_Index + 1;

            elsif
              Src_Tok.Text = Name_R_Paren
              and then Prev_Lexeme (Src_Tokens, Src_Index).Text = Name_R_Paren
               --???Also check that the one before that is a tick!

            then
               Src_Index := Src_Index + 1;

            --  Check for "X : in T" --> "X : T" case

            elsif
              Src_Tok.Normalized = Snames.Name_In
              and then Prev_Lexeme (Src_Tokens, Src_Index).Text = Name_Colon
               --???Check prev&next ids match???
               --  This case should only happen when we're deleting 'in'.

            then
               Src_Index := Src_Index + 1;

            --  ???Deal with empty tasks:

            elsif
              False
              and then -- ???
              Src_Tok.Normalized = Snames.Name_Is
              and then Out_Tok.Text = Name_Semicolon
            then
               raise Program_Error;

            elsif Do_Inserts and then Src_Tok.Kind = End_Of_Line_Comment then
               Insert_End_Of_Line_Comment;

            --  If the source has a blank line at this point, send it to the
            --  output, but avoid multiple blank lines and blank lines just
            --  before End_Of_Input.

            elsif Src_Tok.Kind = Blank_Line then
               loop
                  Src_Index := Src_Index + 1;
                  Src_Tok   := Src_Tokens (Src_Index);
                  exit when Src_Tok.Kind /= Blank_Line;
               end loop;
               if Do_Inserts and then Src_Tok.Kind /= End_Of_Input then
                  Append_Temp_Line_Break;
               end if;

            elsif Do_Inserts and then Src_Tok.Kind = Whole_Line_Comment then
               Insert_Whole_Line_Comment;

            elsif Out_Tok.Kind = Blank_Line then
               Move_Past_Out_Tok;
               Out_Index := Out_Index + 1;

            --  Else print out debugging information and crash. This avoids
            --  damaging the source code in case of bugs.

            else
               Text_IO.Put_Line
                 (Text_IO.Standard_Output,
                  "Token mismatch " & (if Do_Inserts then
                     "(Do_Inserts)"
                  else
                     "(not Do_Inserts)") &
                  ": " &
                  Get_Name_String (Src_Tok.Text) &
                  " --> " &
                  Get_Name_String (Out_Tok.Text));
               Text_IO.Put_Line (Text_IO.Standard_Output, "Src tokens:");
               Put_Tokens
                 (Src_Tokens,
                  First     => Src_Index - 10,
                  Last      => Src_Index + 10,
                  Highlight => Src_Index);
               Text_IO.Put_Line
                 (Text_IO.Standard_Output,
                  "========================================");
               Text_IO.Put_Line (Text_IO.Standard_Output, "Out tokens:");
               Put_Tokens
                 (Out_Tokens,
                  First     => Out_Index - 10,
                  Last      => Out_Index + 10,
                  Highlight => Out_Index);
               raise Program_Error;
            end if;
         end if;
      end loop;

      pragma Assert (Cur_Indentation = 0);

      pragma Assert (Src_Index = Last_Index (Src_Tokens));
      pragma Assert (Out_Index = Last_Index (Out_Tokens));
      pragma Assert (At_End (Out_Buf));
      pragma Assert (Cur_Line = Last_Index (Line_Breaks) + 1);
      pragma Assert (EOL_Cur_Line = Last_Index (EOL_Line_Breaks) + 1);
      pragma Assert (Cur_Tab_Index = Last_Index (Tabs));

      if Do_Inserts then
         pragma Assert (Line_Break_Sorting.Is_Sorted (All_Line_Breaks));
         pragma Assert (Line_Break_Sorting.Is_Sorted (Temp_Line_Breaks));
         Line_Break_Sorting.Merge
           (Target => All_Line_Breaks,
            Source => Temp_Line_Breaks);
      end if;
      pragma Assert (Is_Empty (Temp_Line_Breaks));
      pragma Assert (Line_Break_Sorting.Is_Sorted (All_Line_Breaks));
   end Insert_Comments_And_Blank_Lines;

   procedure Insert_Alignment (Tokens : Scanner.Token_Vectors.Vector);

   procedure Insert_Alignment (Tokens : Scanner.Token_Vectors.Vector) is

      procedure Calculate_Num_Blanks;

      procedure Calculate_Num_Blanks is
         use Scanner;
         Cur_Token_Index : Token_Index := 1;
         function Cur_Tok return Token is (Tokens (Cur_Token_Index));
         Cur_Tab_Index : Tab_Index := 1;
         function Cur_Tab return Tab_Rec is (Tabs (Cur_Tab_Index));

         First_Line_Tabs, Cur_Line_Tabs : Tab_In_Line_Vector;
         --  Tabs for first line of paragraph and for current line.

         First_Line_Indentation, Cur_Line_Indentation : Natural;
         --  Indentation of first line of paragraph and for current line. We
         --  don't consider two lines part of the same paragraph unless their
         --  indentations match, so we don't align the two ":"s in the
         --  following:
         --
         --     procedure P (X : Integer) is
         --        Y : Integer;

         Paragraph_Tabs : Tab_In_Line_Vector_Vectors.Vector;
         --  One Tab_In_Line_Vector for each line in the current paragraph

         procedure Put_Paragraph_Tabs;

         procedure Flush_Para (Up_To : Tab_Index);

         procedure Flush_Para (Up_To : Tab_Index) is -- get rid of param???
            pragma Unreferenced (Up_To);
         begin
            --  Here we have Paragraph_Tabs set to a sequence of lines (or the
            --  tabs in those lines, really). For example, if the input text
            --  was (*1):
            --
            --     package P is
            --
            --        X                    : T                := 1;
            --        A_Long_Variable_Name : T                := 2;
            --        Y                    : A_Long_Type_Name := 3;
            --
            --     end P;
            --     ^
            --     |
            --     column 1
            --
            --  then previous passes will have turned that into (*2):
            --
            --     package P is
            --
            --        X ^1: T ^2:= 1;
            --        A_Long_Variable_Name ^1: T ^2:= 2;
            --        Y ^1: A_Long_Type_Name ^2:= 3;
            --
            --     end P;
            --
            --  The tabs are shown as ^1 and ^2 in (*2) above, although they
            --  are really kept in a separate data structure (Tabs) rather than
            --  in the text itself, and take up zero columns in the buffer. The
            --  "paragraph" we're talking about consists of the three
            --  variable-declaration lines. Note that the alignment from the
            --  input has been forgotten; we would get the same thing if the
            --  input were unaligned. Our job is to align the ":" and ":="
            --  symbols, whether or not they were originally aligned.
            --
            --  ^1 means Index_In_Line = 1; ^2 means Index_In_Line = 2 (see
            --  type Tab_Rec). The Col of each tab is currently set to the
            --  column in which it appears, and the Num_Blanks is currently set
            --  to 0. The following code sets the Col of each tab to the column
            --  in which it WILL appear, and the Num_Blanks to the number of
            --  blanks to expand the tab to in order to achieve that.
            --
            --  We first loop through all the ^1 tabs, and calculate the max
            --  Col, which will be the ":" of the A_Long_Variable_Name line. We
            --  then loop through those again, and set the Num_Blanks to be the
            --  number of blanks needed to reach that max column. For each such
            --  ^1 tab, we loop from that ^1, through ^2 and ^3 and so on (we
            --  have no ^3... in this example), adjusting their Col
            --  accordingly.
            --
            --  Then we loop through all the ^2 tabs in the same way, and so on
            --  for ^3, etc, (if ^3 tabs existed in the example).
            --
            --  So in this example, we loop down through the ^1 tabs to
            --  calculate where to put the ":"'s. Then down through the ^1 tabs
            --  again to adjust the Num_Blanks for the ^1 tabs, and loop across
            --  to adjust the Col for the ^1 and ^2 tabs. Then down through the
            --  ^2 tabs to calculate where to put the ":="'s. Then down through
            --  the ^2 tabs to adjust the Num_Blanks for the ^2 tabs, and loop
            --  across to adjust the Col for the ^2 tabs. Note that adjusting
            --  the Col for the ":"'s affects where we're going to put the
            --  ":="'s -- that's the reason for the "loop across" part.
            --
            --  The end result is to calculate the Num_Blanks so that when we
            --  expand the tabs, (*2) above will be turned (back) into the
            --  (*1).

            --  ???Need to guard against very-long lines in the following.

            if Last_Index (Paragraph_Tabs) = 0 then
               return; -- ???Why needed?
            end if;
            Put_Paragraph_Tabs;

            for Index_In_Line in 1 .. Last_Index (Paragraph_Tabs (1)) loop
               --  ???Should be max line length, if we allow them to differ
               declare
                  Max_Col : Positive := 1;
                  Num_Blanks : Natural;

               begin
                  for Line of Paragraph_Tabs loop
                     declare
                        Tab_I : constant Tab_Index := Line (Index_In_Line);
                        Tab : Tab_Rec renames Tabs (Tab_I);

                     begin
                        Max_Col := Positive'Max (Max_Col, Tab.Col);
                     end;
                  end loop;

                  for Line of Paragraph_Tabs loop
                     declare
                        Tab_I : constant Tab_Index := Line (Index_In_Line);
                        Tab : Tab_Rec renames Tabs (Tab_I);

                     begin
                        Num_Blanks     := Max_Col - Tab.Col;
                        Tab.Num_Blanks := Num_Blanks;

                        for X_In_Line in Index_In_Line .. Last_Index (Line)
                        loop
                           declare
                              Tab_J : constant Tab_Index := Line (X_In_Line);
                              Tab_2 : Tab_Rec renames Tabs (Tab_J);

                           begin
                              Tab_2.Col := Tab_2.Col + Num_Blanks;
                           end;
                        end loop;
                        pragma Assert (Tab.Col = Max_Col);
                     end;
                  end loop;
               end;
            end loop;
            Put_Paragraph_Tabs;

            Clear (Paragraph_Tabs);
         end Flush_Para;

         procedure Process_Line;

         procedure Process_Line is
            Lexeme_Seen_On_Line : Boolean := False;
            pragma Unreferenced (Lexeme_Seen_On_Line); -- ???
            Tab_Mismatch        : Boolean := False;

         begin
            Cur_Line_Indentation := Cur_Tok.Sloc.Col - 1;

            while Cur_Tok.Kind not in End_Of_Input | End_Of_Line | Blank_Line
            loop
               if
                 Cur_Tok.Kind not in
                   Start_Of_Input |
                     Whole_Line_Comment |
                     End_Of_Line_Comment
               then
                  Lexeme_Seen_On_Line := True;
               end if;

               pragma Assert
                 (Cur_Tok.Sloc.First <= Position (Out_Buf, Cur_Tab.Mark));
               if Cur_Tok.Sloc.Firstx = Cur_Tab.Mark then
                  if not Tab_Mismatch then
                     if
                       Cur_Tab.Index_In_Line = Last_Index (Cur_Line_Tabs) + 1
                     then
                        Append (Cur_Line_Tabs, Cur_Tab_Index);

                     else
                        Tab_Mismatch := True;
                     end if;
                  end if;

                  Tabs (Cur_Tab_Index).Col := Cur_Tok.Sloc.Col;
                  Cur_Tab_Index            := Cur_Tab_Index + 1;
               end if;

               Cur_Token_Index := Cur_Token_Index + 1;
            end loop;

            if Tab_Mismatch then
               Clear (Cur_Line_Tabs);
            end if;
         end Process_Line;

         function Tokens_Match (X, Y : Tab_In_Line_Vector) return Boolean;

         function Tokens_Match (X, Y : Tab_In_Line_Vector) return Boolean is
         begin
            pragma Assert (not Is_Empty (Y));
            if Last_Index (X) /= Last_Index (Y) then
               return False;
            end if;

            for J in 1 .. Last_Index (X) loop
               declare
                  XX : constant Tab_Index := X (J);
                  YY : constant Tab_Index := Y (J);

               begin
                  if Tabs (XX).Token /= Tabs (YY).Token then
                     return False;
                  end if;
               end;
            end loop;
            return True;
         end Tokens_Match;

         procedure Put_Tab_In_Line_Vector
           (Name : String;
            X    : Tab_In_Line_Vector);

         procedure Put_Tab_In_Line_Vector
           (Name : String;
            X    : Tab_In_Line_Vector)
         is
         begin
            if Is_Empty (X) then
               return;
            end if;

            Dbg_Out.Put ("\1: \t", Name);

            for J in 1 .. Last_Index (X) loop
               if J /= 1 then
                  Dbg_Out.Put ("; ");
               end if;
               Dbg_Out.Put ("\1", Tab_Image (X (J)));
            end loop;
            Dbg_Out.Put ("\n");
         end Put_Tab_In_Line_Vector;

         procedure Put_Paragraph_Tabs is
         begin
            Dbg_Out.Put
              ("\1 Paragraph_Tabs\n",
               Image (Integer (Last_Index (Paragraph_Tabs))));

            for X of Paragraph_Tabs loop
               Put_Tab_In_Line_Vector ("", X);
            end loop;
            Dbg_Out.Put ("end Paragraph_Tabs\n");
         end Put_Paragraph_Tabs;

      --  Start of processing for Calculate_Num_Blanks

      begin
         while Cur_Tok.Kind /= End_Of_Input loop
            declare
               First_Char_In_Line : constant Natural :=
                 Cur_Tok.Sloc.First - Cur_Tok.Sloc.Col + 1;
               Saved_Cur_Tab_Index : constant Tab_Index := Cur_Tab_Index;

            begin
               Process_Line;

               Dbg_Out.Put ("<<");

               for X in First_Char_In_Line .. Cur_Tok.Sloc.First - 1 loop
                  for Tab of Cur_Line_Tabs loop
                     if X = Position (Out_Buf, Tabs (Tab).Mark) then
                        Dbg_Out.Put ("^");
                     end if;
                  end loop;
                  Dbg_Out.Put_Char (Char_At (Out_Buf, X));
               end loop;
               Dbg_Out.Put (">>\n");
               Put_Tab_In_Line_Vector ("First", First_Line_Tabs);
               Put_Tab_In_Line_Vector ("Cur", Cur_Line_Tabs);

               Cur_Token_Index := Cur_Token_Index + 1;
               --  Consume the newline

               if Is_Empty (Cur_Line_Tabs) then
                  Dbg_Out.Put ("Flush_Para -- no tabs\n");
                  Flush_Para (Saved_Cur_Tab_Index);
                  --  Leave tabs from this line with Num_Blanks = 0.
                  Clear (First_Line_Tabs);
                  First_Line_Indentation := Natural'Last;

               else
                  if Is_Empty (First_Line_Tabs) then
                     First_Line_Tabs        := Cur_Line_Tabs;
                     First_Line_Indentation := Cur_Line_Indentation;

                  else
                     --  If the tokens don't match, we're at the end of a
                     --  paragraph. If the indentation doesn't match, we're
                     --  also at the end of a paragraph. We allow the
                     --  indentation to be off by one and still match, to deal
                     --  with things like:
                     --
                     --     procedure P
                     --       (Param         : T;
                     --        Another_Param : T);

                     if
                       not Tokens_Match (Cur_Line_Tabs, First_Line_Tabs)
                       or else Cur_Line_Indentation not in
                         First_Line_Indentation .. First_Line_Indentation + 1
                     then
                        Dbg_Out.Put ("Flush_Para -- token mismatch\n");
                        Flush_Para (Saved_Cur_Tab_Index);
                        First_Line_Tabs        := Cur_Line_Tabs;
                        First_Line_Indentation := Cur_Line_Indentation;
                     end if;
                  end if;
                  Append (Paragraph_Tabs, Cur_Line_Tabs);
                  Clear (Cur_Line_Tabs);
               end if;
            end;
            Dbg_Out.Put ("\n");
         end loop;

         pragma Assert (Cur_Tab_Index = Last_Index (Tabs));
      end Calculate_Num_Blanks;

   --  Start of processing for Insert_Alignment

   begin
      --  First go through the tabs and set their Num_Blanks field to the
      --  appropriate value. Tabs that are not expanded at all will have
      --  Num_Blanks left equal to zero.

      Format_Debug_Output ("before Calculate_Num_Blanks");
      Calculate_Num_Blanks;
      Format_Debug_Output ("after Calculate_Num_Blanks");

      --  Now go through the buffer, inserting blanks for tabs that should be
      --  expanded. Don't expand a tab if it would make the line too long.

      declare
         Cur_Tab_Index : Tab_Index := 1;
         Cur_Tab       : Tab_Rec   := Tabs (Cur_Tab_Index);
         Cur_Line_Num  : Positive  := 1;

      begin
         while not At_End (Out_Buf) loop
            pragma Assert
              (Point (Out_Buf) <= Position (Out_Buf, Cur_Tab.Mark));

            while At_Point (Out_Buf, Cur_Tab.Mark) loop
               if
                 Scanner.Line_Length
                   (Out_Buf,
                    Out_Buf_Line_Ends,
                    Cur_Line_Num) +
                 Cur_Tab.Num_Blanks <=
                 Options.Max_Line_Length
               then
                  for J in 1 .. Cur_Tab.Num_Blanks loop
                     Insert (Out_Buf, ' ');
                  end loop;
               end if;
               Cur_Tab_Index := Cur_Tab_Index + 1;
               Cur_Tab       := Tabs (Cur_Tab_Index);
            end loop;
            if Cur (Out_Buf) = NL then
               Cur_Line_Num := Cur_Line_Num + 1;
            end if;
            Move_Forward (Out_Buf);
         end loop;
         pragma Assert (Cur_Tab_Index = Last_Index (Tabs));
      end;
   end Insert_Alignment;

--  Start of processing for Tree_To_Ada

begin
   --  Note that if we're processing multiple files, we will get here multiple
   --  times, so we need to clear out data structures left over from last
   --  time.

   pragma Assert (Cur_Indentation = 0);
   Clear (All_Line_Breaks);
   pragma Assert (Is_Empty (Line_Stack));
   Clear (Tabs);

   Clear (Source_Buf);
   Insert (Source_Buf, Source_Text);
   Reset (Source_Buf);
   pragma Assert (To_String (Source_Buf) = Source_Text);
   Scanner.Get_Tokens (Source_Buf, Source_Tokens);
   if Debug_Mode then
      Dbg_Out.Put ("Source_Tokens:\n");
      Scanner.Put_Tokens (Source_Tokens);
      Dbg_Out.Put ("end Source_Tokens:\n");
   end if;

   Clear (Out_Buf);

   Append_Line_Break
     (Hard     => True,
      Level    => 0,
      Kind     => Not_An_Element,
      Template => Name_Empty);
   Check_Whitespace := True;
   Subtree_To_Ada (Tree, Cur_Level => 0, Index_In_Parent => 1);
   Check_Whitespace := False;
   Append (Tabs, Tab_Rec'(Mark => Mark (Out_Buf, '$'), others => <>));
   --  Append a sentinel tab, whose Position is greater than any actual
   --  position. This ensures that as we step through Tabs, there is always one
   --  more.
   pragma Assert (Is_Empty (Tree_Stack));
   pragma Assert (not L_Square_Seen);

   Format_Debug_Output ("before Split_Lines");
   Split_Lines;

   Format_Debug_Output ("before Insert_Comments_And_Blank_Lines");
   Reset (Out_Buf);
   Scanner.Get_Tokens (Out_Buf, Out_Tokens);
   --  ???At this point, we need another pass to insert hard line breaks after
   --  end-of-line comments, so they will be indented properly. Or better yet,
   --  insert the EOL comments, with tabs and soft line break before, hard line
   --  break after.
   Reset (Out_Buf);
   Insert_Comments_And_Blank_Lines
     (Src_Tokens => Source_Tokens,
      Out_Tokens => Out_Tokens,
      Do_Inserts => True);
   pragma Assert (At_End (Out_Buf) and then Lookback (Out_Buf) = NL);

   Clear (Out_Tokens);

   Format_Debug_Output ("before Split_Lines again");
   Split_Lines;
   Format_Debug_Output ("after Split_Lines again");
   pragma Assert (Point (Out_Buf) = Last_Position (Out_Buf) + 1);
   pragma Assert
     (Position (Out_Buf, All_Line_Breaks (Last (All_Line_Breaks)).Mark) =
      Last_Position (Out_Buf));
   pragma Assert (At_End (Out_Buf));

   Reset (Out_Buf);
   Dbg_Out.Put ("Insert_NLs_And_Indentation input:\n");
   Dump_Buffer (Out_Buf);
   Insert_NLs_And_Indentation;
   Dbg_Out.Put ("Insert_NLs_And_Indentation output:\n");
   Dump_Buffer (Out_Buf);

   Reset (Out_Buf);
   Scanner.Get_Tokens (Out_Buf, Out_Tokens);

   pragma Assert (Char_At (Out_Buf, Last_Position (Out_Buf)) = NL);
   Move (Target => Temp_Tokens, Source => Out_Tokens);
   Reset (Out_Buf);
   Clear (Out_Buf_Line_Ends);
   Scanner.Get_Tokens
     (Out_Buf,
      Out_Tokens,
      Ignore_Single_Line_Breaks => False,
      Line_Ends                 => Out_Buf_Line_Ends'Access);
   Scanner.Check_Same_Tokens (Temp_Tokens, Out_Tokens);
   Reset (Out_Buf);
   Insert_Alignment (Tokens => Out_Tokens);

   Move (Target => Temp_Tokens, Source => Out_Tokens);
   Reset (Out_Buf);
   Scanner.Get_Tokens (Out_Buf, Out_Tokens);
   Scanner.Check_Same_Tokens (Temp_Tokens, Out_Tokens);
   Clear (Temp_Tokens);

   --  The following pass doesn't modify anything; it just checks that the
   --  sequence of tokens we have constructed matches the original source code
   --  (with some allowed exceptions).

   Reset (Out_Buf);
   Insert_Comments_And_Blank_Lines
     (Src_Tokens => Source_Tokens,
      Out_Tokens => Out_Tokens,
      Do_Inserts => False);

   Reset (Out_Buf);
   Write_Buffer (Out_Buf);
exception
   when X : others =>
      Text_IO.Set_Output (Text_IO.Standard_Output);
      Text_IO.Put_Line (Ada.Exceptions.Exception_Information (X));

      while not At_End (Out_Buf) loop
         Move_Forward (Out_Buf);
      end loop;
      Insert (Out_Buf, NL);
      Reset (Out_Buf);
      Write_Buffer (Out_Buf);
      Debug_Mode := True;
      Format_Debug_Output ("unhandled exception");
      raise;
end Tree_To_Ada;
