------------------------------------------------------------------------------
--                                                                          --
--                            GNAT2XML COMPONENTS                           --
--                                                                          --
--                      G N A T 2 X M L . S C A N N E R                     --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                  Copyright (C) 2012-2013, AdaCore, Inc.                  --
--                                                                          --
-- Gnat2xml is free software; you can redistribute it and/or modify it      --
-- under terms of the  GNU General Public License  as published by the Free --
-- Software Foundation;  either version 2,  or  (at your option)  any later --
-- version. Gnat2xml is distributed  in the hope  that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of MER-      --
-- CHANTABILITY or  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General  --
-- Public License for more details. You should have received a copy of the  --
-- GNU General Public License distributed with GNAT; see file COPYING. If   --
-- not, write to the Free Software Foundation, 59 Temple Place Suite 330,   --
-- Boston, MA 02111-1307, USA.                                              --
-- The gnat2xml tool was derived from the Avatox sources.                   --
------------------------------------------------------------------------------

pragma Ada_2012;

with Ada.Characters.Handling; use Ada.Characters.Handling;
use Ada.Characters;
with Text_IO;

with Opt;
with Snames;
with Scans;

package body Ada_Trees.Scanner is

   package Char_Sets is

      --  Define some interesting character sets as subtypes of Character

      subtype Line_Ending_Char is Character with
           Predicate => Line_Ending_Char in LF | CR | VT | FF;

      subtype Whitespace is Character with
           Predicate => Whitespace in ' ' | HT | LF | CR | VT | FF;
      pragma Unreferenced (Whitespace);

      subtype Initial_Identifier_Char is Character with
           Predicate =>
             Initial_Identifier_Char in
               'A' .. 'Z' |
                 'a' .. 'z' |
                 '_' |
                 UC_A_Grave .. UC_O_Diaeresis |
                 UC_O_Oblique_Stroke .. UC_Icelandic_Thorn |
                 LC_German_Sharp_S .. LC_O_Diaeresis |
                 LC_O_Oblique_Stroke .. LC_Y_Diaeresis;
         --  Characters that can be the initial character in an identifier
      pragma Unreferenced (Initial_Identifier_Char);

      subtype Identifier_Char is Character with
           Predicate =>
             Identifier_Char in
               'A' .. 'Z' |
                 'a' .. 'z' |
                 '0' .. '9' |
                 '_' |
                 UC_A_Grave .. UC_O_Diaeresis |
                 UC_O_Oblique_Stroke .. UC_Icelandic_Thorn |
                 LC_German_Sharp_S .. LC_O_Diaeresis |
                 LC_O_Oblique_Stroke .. LC_Y_Diaeresis;
         --  Characters that can be the initial or subsequent character in an
         --  identifier

      subtype Based_Digit_Char is Character with
           Predicate =>
             Based_Digit_Char in 'A' .. 'F' | 'a' .. 'f' | '0' .. '9' | '_';
      pragma Unreferenced (Based_Digit_Char);

      subtype Decimal_Digit_Char is Character with
           Predicate => Decimal_Digit_Char in '0' .. '9' | '_';

   end Char_Sets;

   use Char_Sets;

   function First_Pos
     (Input : Buffer;
      Sloc  : Source_Location)
     return Positive
   is
   begin
      return Result : constant Positive := Position (Input, Sloc.Firstx) do
         pragma Assert (Result = Sloc.First);
      end return;
   end First_Pos;

   function Last_Pos
     (Input : Buffer;
      Sloc  : Source_Location)
     return Positive
   is
   begin
      return Result : constant Positive := Position (Input, Sloc.Lastx) do
         pragma Assert (Result = Sloc.Last);
      end return;
   end Last_Pos;

   procedure Get_Tokens
     (Input  : in out Buffer;
      Result : out Token_Vectors.Vector;
      Ignore_Single_Line_Breaks : Boolean := True;
      Max_Tokens : Token_Index := Token_Index'Last;
      Line_Ends : Marker_Vector_Ptr := null)
   is
      pragma Assert (if Line_Ends /= null then
            Char_At (Input, Last_Position (Input)) = NL);

      Name_Empty   : constant Name_Id := Name_Find ("");
      Name_R_Paren : constant Name_Id := Name_Find (")");
      Name_NL : constant Name_Id := Name_Find ((1 => NL));

      Cur_Line, Cur_Col : Positive := 1;
      Cur_First         : Positive := 1;

      procedure Get;
      --  Move ahead one character in the input

      procedure Get is
      begin
         Cur_Col := Cur_Col + 1;

         Cur_First := Cur_First + 1;
         --  It's OK for this to be greater than Input'Last; in that case,
         --  Cur will return Latin_1.NUL, indicating end-of-input.

         Move_Forward (Input);
      end Get;

      function Get_Token return Token;
      --  Get one token from the input

      procedure Put;
      --  Append Cur (Input) onto the end of the Name_Buffer to be later
      --  captured by Name_Find.

      procedure Scan_Decimal_Digit_Chars;
      procedure Scan_Identifier_Chars;

      procedure Scan_String_Literal (Q_Char : Character);
      --  Scan out a string literal, where Q_Char is the initial quote
      --  character (either '"' or '%'), taking care of doubled quote
      --  characters.

      procedure Scan_Decimal_Digit_Chars is
      begin
         while Cur (Input) in Decimal_Digit_Char loop
            Put;
            Get;
         end loop;
      end Scan_Decimal_Digit_Chars;

      procedure Scan_Identifier_Chars is
      begin
         while Cur (Input) in Identifier_Char loop
            Put;
            Get;
         end loop;
      end Scan_Identifier_Chars;

      procedure Scan_String_Literal (Q_Char : Character) is
      begin
         loop
            Put;
            Get;

            if Cur (Input) = Latin_1.NUL then
               raise Program_Error with "end of file in string literal";
            end if;

            if Cur (Input) = Q_Char then
               Put;
               Get;

               exit when Cur (Input) /= Q_Char;
            end if;
         end loop;
      end Scan_String_Literal;

      Preceding_Lexeme, Preceding_Token : Token := (Kind => Nil, others => <>);
      --  These record the preceding token, which is necessary to
      --  properly deal with single quotes and line breaks.
      --  Preceding_Lexeme doesn't count comments and line breaks;
      --  Preceding_Token does.

      function Get_Token return Token is
      begin
         while Cur (Input) in ' ' | HT loop
            Get;
         end loop;

         Name_Len := 0;

         return Result : Token do
            Result.Sloc :=
              (Line  => Cur_Line,
              Col    => Cur_Col,
              First  => Cur_First,
              Last   => <>,
              Firstx => Mark (Input, '<'),
              Lastx  => <>);

            case Cur (Input) is
               when Latin_1.NUL =>
                  Result.Kind := End_Of_Input;

               when ' ' | HT =>
                  raise Program_Error;

               when LF | CR | VT | FF =>
                  if Line_Ends /= null then
                     Append (Line_Ends.all, Result.Sloc.Firstx);
                  end if;

                  if Cur (Input) = CR and then Lookahead (Input) = LF then
                     pragma Assert (Line_Ends = null);
                     Put;
                     Get;
                  end if;

                  Put;
                  Get;
                  Result.Kind := End_Of_Line;
                  --  Might be ignored below, or turned into Blank_Line

                  Cur_Line := Cur_Line + 1;
                  Cur_Col  := 1;

               --  Minus sign or comment

               when '-' =>
                  Put;
                  Get;

                  if Cur (Input) = '-' then
                     if Preceding_Lexeme.Sloc.Line = Cur_Line then
                        Result.Kind := End_Of_Line_Comment;

                     else
                        Result.Kind := Whole_Line_Comment;
                     end if;

                     loop
                        Put;
                        Get;

                        exit when Cur (Input) in
                          Line_Ending_Char |
                            Latin_1.NUL;
                     end loop;

                  else
                     Result.Kind := Lexeme;
                  end if;

               --  identifier

--???               when Initial_Identifier_Char =>

               when
                 'A' .. 'Z' |
                 'a' .. 'z' |
                 '_' |
                 UC_A_Grave .. UC_O_Diaeresis |
                 UC_O_Oblique_Stroke .. UC_Icelandic_Thorn |
                 LC_German_Sharp_S .. LC_O_Diaeresis |
                 LC_O_Oblique_Stroke .. LC_Y_Diaeresis =>
                  Result.Kind                          := Identifier;
                  --  We will check for reserved words below
                  Scan_Identifier_Chars;

               --  numeric literal

               when '0' .. '9' =>
                  Result.Kind := Lexeme;
                  Scan_Decimal_Digit_Chars;

                  if Cur (Input) in '#' | ':' then
                     loop
                        Put;
                        Get;

                        exit when Cur (Input) in '#' | ':' | Latin_1.NUL;
                     end loop;

                     Put;
                     Get;

                  elsif Cur (Input) = '.' then
                     if Lookahead (Input) /= '.' then -- could be ".."
                        Put;
                        Get;
                        Scan_Decimal_Digit_Chars;
                     end if;
                  end if;

                  if To_Lower (Cur (Input)) = 'e' then
                     Put;
                     Get;

                     if Cur (Input) in '+' | '-' then
                        Put;
                        Get;
                     end if;

                     Scan_Decimal_Digit_Chars;
                  end if;

               --  single quote or character literal

               when ''' =>
                  Result.Kind := Lexeme;
                  Put;
                  Get;

                  --  We distinguish a single quote token from a character
                  --  literal by looking back at the preceding token. If it's
                  --  not one of the following, the it's a single quote. For
                  --  example, in Character'('''), the first quote follows an
                  --  identifier, so it's a single quote. The second one
                  --  follows a left parenthesis, so it's the start of a
                  --  character literal.

                  if
                    Preceding_Lexeme.Kind /= Identifier
                    and then Preceding_Lexeme.Normalized /= Snames.Name_Access
                    and then Preceding_Lexeme.Normalized /= Snames.Name_All
                    and then Preceding_Lexeme.Normalized /= Name_R_Paren
                  then
                     Put;
                     Get;
                     pragma Assert (Cur (Input) = ''');
                     Put;
                     Get;
                  end if;

               --  string literal

               when '"' | '%' =>
                  Result.Kind                 := String_Literal;
                  Scan_String_Literal (Q_Char => Cur (Input));

               --  One-character tokens

               when '&' | '(' | ')' | '+' | ',' | ';' | '|' | '!' =>
                  Result.Kind := Lexeme;
                  Put;
                  Get;

               --  Multiple-character tokens. We need to distinguish between
               --  "=" and "=>", and between "." and ".." and so forth.

               when '=' =>
                  Result.Kind := Lexeme;
                  Put;
                  Get;

                  if Cur (Input) = '>' then
                     Put;
                     Get;
                  end if;

               when '.' =>
                  Result.Kind := Lexeme;
                  Put;
                  Get;

                  if Cur (Input) = '.' then
                     Put;
                     Get;
                  end if;

               when '*' =>
                  Result.Kind := Lexeme;
                  Put;
                  Get;

                  if Cur (Input) = '*' then
                     Put;
                     Get;
                  end if;

               when '/' =>
                  Result.Kind := Lexeme;
                  Put;
                  Get;

                  if Cur (Input) = '=' then
                     Put;
                     Get;
                  end if;

               when '>' =>
                  Result.Kind := Lexeme;
                  Put;
                  Get;

                  if Cur (Input) in '=' | '>' then
                     Put;
                     Get;
                  end if;

               when '<' =>
                  Result.Kind := Lexeme;
                  Put;
                  Get;

                  if Cur (Input) in '=' | '<' | '>' then
                     Put;
                     Get;
                  end if;

               when ':' =>
                  Result.Kind := Lexeme;
                  Put;
                  Get;

                  if Cur (Input) = '=' then
                     Put;
                     Get;
                  end if;

               when others =>
                  raise Program_Error
                    with "unrecognized character: " & Cur (Input);
                  --  All legal token-starting characters are handled above
            end case;

            pragma Assert (Result.Kind /= Nil);
            Result.Text := Name_Find;
            case Result.Kind is
               when End_Of_Line | Blank_Line =>
                  Result.Normalized := Name_NL;

               when Whole_Line_Comment | End_Of_Line_Comment =>
                  --  Remove trailing whitespace

                  while Name_Buffer (Name_Len) in ' ' | HT loop
                     Name_Len := Name_Len - 1;
                  end loop;
                  Result.Normalized := Name_Find;

               when others =>
                  Name_Buffer (1 .. Name_Len) :=
                    To_Lower (Name_Buffer (1 .. Name_Len));
                  Result.Normalized := Name_Find;
            end case;

            if -- Check for reserved word
              Result.Kind = Identifier
              and then Snames.Is_Keyword_Name (Result.Normalized)
            then
               Result.Kind := Reserved_Word;
            end if;

            Result.Sloc.Lastx := Mark (Input, '>');
            Result.Sloc.Last  := Cur_First - 1;
         end return;
      end Get_Token;

      procedure Put is
      begin
         Add_Char_To_Name_Buffer (Cur (Input));
      end Put;

      Start_Mark  : constant Marker := Mark (Input, '!');
      Start_Token : Token           :=
        (Start_Of_Input,
         Text | Normalized => Name_Empty,
         Sloc =>
           (Line => Cur_Line,
           Col   => Cur_Col,
           First => Cur_First,
           Last  => 0,
           Firstx | Lastx => Start_Mark));

   --  Start of processing for Get_Tokens

   begin
      pragma Assert (Point (Input) = 1);
      Clear (Result);
      Start_Token.Sloc.Lastx := Mark (Input, '>');
      Append (Result, Start_Token);

      if Line_Ends /= null then
         Clear (Line_Ends.all);
      end if;

      loop
         exit when Result.Last_Index - 1 = Max_Tokens;
         --  "- 1" because Start_Of_Input doesn't count

         declare
            T : Token := Get_Token;

         begin
            pragma Assert
              (Get_Name_String (T.Text) =
               Slice (Input, T.Sloc.Firstx, T.Sloc.Lastx));

            if T.Kind = End_Of_Line then
               if Preceding_Token.Kind in End_Of_Line | Blank_Line then
                  T.Kind := Blank_Line;

               elsif Ignore_Single_Line_Breaks then
                  goto Ignore_It;
               end if;
            end if;

            Append (Result, T);
            <<Ignore_It>>

            exit when T.Kind = End_Of_Input;

            Preceding_Token := T; -- even if an ignored End_Of_Line
            if T.Kind not in Blank_Line | End_Of_Line | Comment_Kind then
               Preceding_Lexeme := T;
            end if;
         end;
      end loop;

      if Line_Ends /= null then
         for E in 1 .. Last_Index (Line_Ends.all) loop
            pragma Assert
              (Char_At (Input, Position (Input, Line_Ends.all (E))) = NL);
         end loop;
      end if;
   end Get_Tokens;

   function Get_Token (Input : String) return Token is
      Tokens : Token_Vectors.Vector;
      Buf : Buffer := String_To_Buffer (Input);

   begin
      Get_Tokens
        (Buf,
         Tokens,
         Ignore_Single_Line_Breaks => True,
         Max_Tokens                => 1);
      pragma Assert (Tokens (1).Kind = Start_Of_Input);
      pragma Assert (Last_Index (Tokens) = 2);
      return Tokens (2);
   end Get_Token;

   function Line_Length
     (Input    : in out Buffer;
      Ends     : Marker_Vector;
      Line_Num : Positive)
     return Natural
   is

      M1 : constant Marker   := Ends (Marker_Index (Line_Num));
      P1 : constant Positive := Position (Input, M1);
      M0 : Marker;
      P0 : Natural;

   begin
      if Line_Num = 1 then
         P0 := 0;

      else
         M0 := Ends (Marker_Index (Line_Num - 1));
         P0 := Position (Input, M0);
      end if;
      return P1 - P0 - 1;
   end Line_Length;

   procedure Put_Token (Tok : Token; Index : Token_Index := 1) is
   begin
      Text_IO.Put_Line
        (Text_IO.Standard_Output,
         Get_Name_String (Tok.Text) &
         " -- #" &
         Image (Integer (Index)) &
         "  " &
         Capitalize (Tok.Kind'Img) &
         " at " &
         Image (Tok.Sloc));
   end Put_Token;

   procedure Put_Tokens
     (Tokens : Token_Vectors.Vector;
      First : Token_Index'Base := 1;
      Last : Token_Index'Base := Token_Index'Last;
      Highlight : Token_Index'Base := 0)
   is
   begin
      for Index in
        Token_Index'Max (First, 1) ..
            Token_Index'Min (Last, Last_Index (Tokens))
      loop
         if Index in 1 .. Last_Index (Tokens) then
            if Index = Highlight then
               Text_IO.Put_Line (Text_IO.Standard_Output, "----------------");
            end if;

            Put_Token (Tokens (Index), Index);
         end if;
      end loop;
   end Put_Tokens;

   procedure Check_Same_Tokens (X, Y : Token_Vectors.Vector) is
      Xj, Yj : Token_Index := 1;

   begin
      loop
         declare
            X_Tok  : constant Token   := X (Xj);
            Y_Tok  : constant Token   := Y (Yj);
            X_Done : constant Boolean := X_Tok.Kind = End_Of_Input;
            Y_Done : constant Boolean := Y_Tok.Kind = End_Of_Input;

         begin
            if Xj = 1 then
               pragma Assert (X_Tok.Kind = Start_Of_Input);
               pragma Assert (Y_Tok.Kind = Start_Of_Input);
            end if;

            if X_Tok.Kind in End_Of_Line | Blank_Line then
               Xj := Xj + 1;
               goto Continue;
            end if;
            if Y_Tok.Kind in End_Of_Line | Blank_Line then
               Yj := Yj + 1;
               goto Continue;
            end if;

            pragma Assert (X_Tok.Kind = Y_Tok.Kind);
            pragma Assert (X_Tok.Text = Y_Tok.Text);
            pragma Assert (X_Tok.Normalized = Y_Tok.Normalized);
            --  Ignore Sloc

            pragma Assert (X_Done = (Xj = Last_Index (X)));
            pragma Assert (Y_Done = (Yj = Last_Index (Y)));
            pragma Assert (X_Done = Y_Done);
            exit when X_Done;

            Xj := Xj + 1;
            Yj := Yj + 1;
            <<Continue>>

         end;
      end loop;
   end Check_Same_Tokens;

begin
   --  Initialize things so we can call Snames.Is_Keyword_Name. Use Ada_2012 so
   --  it will recognize the newer reserved words. ???Actually, we really ought
   --  to use the actual Ada_Version in use, but that's difficult to
   --  determine. It's harmless to use Ada 95 for now, because there's no
   --  special casing of the newer reserved words.

   if False then
      Opt.Ada_Version := Opt.Ada_2012;
   end if;

   Snames.Initialize;
   Scans.Initialize_Ada_Keywords;
end Ada_Trees.Scanner;
